import logging
from typing import Optional

from r2r.core import EmbeddingProvider

logger = logging.getLogger(__name__)


class SentenceTransformerEmbeddingProvider(EmbeddingProvider):
    def __init__(
        self, embedding_model: str, provider: str = "sentence-transformers"
    ):
        logger.info(
            "Initializing `SentenceTransformerEmbeddingProvider` to provide embeddings."
        )

        super().__init__(provider)
        if provider != "sentence-transformers":
            raise ValueError(
                "SentenceTransformerEmbeddingProvider must be initialized with provider `sentence-transformers`."
            )
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            raise ValueError(
                "Must download sentence-transformers library to run `SentenceTransformerEmbeddingProvider`."
            )
        self.encoder = SentenceTransformer(embedding_model)

    def _check_inputs(self, model: str, dimensions: Optional[int]) -> None:
        if (
            dimensions
            and dimensions != self.encoder.get_sentence_embedding_dimension()
        ):
            raise ValueError(
                f"Dimensions {dimensions} for {model} are not supported"
            )

    def get_embedding(
        self, text: str, model: str, dimensions: Optional[int] = None
    ) -> list[float]:
        self._check_inputs(model, dimensions)
        return self.encoder.encode([text]).tolist()[0]

    def get_embeddings(
        self, texts: list[str], model: str, dimensions: Optional[int] = None
    ) -> list[list[float]]:
        self._check_inputs(model, dimensions)
        return self.encoder.encode(texts).tolist()

    def tokenize_string(self, text: str, model: str) -> list[int]:
        raise ValueError(
            "SentenceTransformerEmbeddingProvider does not support `tokenize_string`."
        )



import logging
import os
from typing import Optional, Union

from r2r.core import VectorDBProvider, VectorEntry, VectorSearchResult
from r2r.vecs.client import Client
from r2r.vecs.collection import Collection

logger = logging.getLogger(__name__)


class PGVectorDB(VectorDBProvider):
    def __init__(self, provider: str = "pgvector") -> None:
        logger.info(
            "Initializing `PGVectorDB` to store and retrieve embeddings."
        )

        super().__init__(provider)
        if provider != "pgvector":
            raise ValueError(
                "PGVectorDB must be initialized with provider `pgvector`."
            )
        try:
            import r2r.vecs
        except ImportError:
            raise ValueError(
                f"Error, PGVectorDB requires the vecs library. Please run `poetry add vecs`."
            )
        try:
            user = os.getenv("POSTGRES_USER")
            password = os.getenv("POSTGRES_PASSWORD")
            host = os.getenv("POSTGRES_HOST")
            port = os.getenv("POSTGRES_PORT")
            db_name = os.getenv("POSTGRES_DBNAME")

            DB_CONNECTION = (
                f"postgresql://{user}:{password}@{host}:{port}/{db_name}"
            )
            self.vx: Client = r2r.vecs.create_client(DB_CONNECTION)
        except Exception as e:
            raise ValueError(
                f"Error {e} occurred while attempting to connect to the pgvector provider with {DB_CONNECTION}."
            )
        self.collection: Optional[Collection] = None

    def initialize_collection(
        self, collection_name: str, dimension: int
    ) -> None:
        self.collection = self.vx.get_or_create_collection(
            name=collection_name, dimension=dimension
        )

    def copy(self, entry: VectorEntry, commit=True) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `copy`."
            )

        self.collection.copy(
            records=[(str(entry.id), entry.vector, entry.metadata)]
        )

    def copy_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `copy_entries`."
            )

        self.collection.copy(
            records=[
                (str(entry.id), entry.vector, entry.metadata)
                for entry in entries
            ]
        )

    def upsert(self, entry: VectorEntry, commit=True) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert`."
            )

        self.collection.upsert(
            records=[(str(entry.id), entry.vector, entry.metadata)]
        )

    def upsert_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert_entries`."
            )

        self.collection.upsert(
            records=[
                (str(entry.id), entry.vector, entry.metadata)
                for entry in entries
            ]
        )

    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `search`."
            )
        measure = kwargs.get("measure", "cosine_distance")
        mapped_filters = {
            key: {"$eq": value} for key, value in filters.items()
        }

        return [
            VectorSearchResult(ele[0], 1 - ele[1], ele[2])  # type: ignore
            for ele in self.collection.query(
                data=query_vector,
                limit=limit,
                filters=mapped_filters,
                measure=measure,
                include_value=True,
                include_metadata=True,
            )
        ]

    def create_index(self, index_type, column_name, index_options):
        pass

    def close(self):
        pass

    def filtered_deletion(
        self, key: str, value: Union[bool, int, str]
    ) -> None:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `filtered_deletion`."
            )
        self.collection.delete(filters={key: {"$eq": value}})  # type: ignore

    def get_all_unique_values(
        self, metadata_field: str, filters: dict = {}
    ) -> list:
        if self.collection is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `get_all_unique_values`."
            )

        mapped_filters = {
            key: {"$eq": value} for key, value in filters.items()
        }

        # Pass an empty vector as the `data` argument
        records = self.collection.query(
            data=[0] * self.collection.dimension,  # Empty vector
            filters=mapped_filters,
            include_metadata=True,
            include_value=False,
            # Remove the `limit` argument to retrieve all records
        )

        unique_values = set(
            record[1].get(metadata_field) for record in records
        )

        return list(unique_values)



import logging
import os
from typing import Optional

from openai import OpenAI

from r2r.core import EmbeddingProvider

logger = logging.getLogger(__name__)


class OpenAIEmbeddingProvider(EmbeddingProvider):
    MODEL_TO_TOKENIZER = {
        "text-embedding-ada-002": "cl100k_base",
        "text-embedding-3-small": "cl100k_base",
        "text-embedding-3-large": "cl100k_base",
    }
    MODEL_TO_DIMENSIONS = {
        "text-embedding-ada-002": [1536],
        "text-embedding-3-small": [512, 1536],
        "text-embedding-3-large": [256, 1024, 3072],
    }

    def __init__(self, provider: str = "openai"):
        logger.info(
            "Initializing `OpenAIEmbeddingProvider` to provide embeddings."
        )

        super().__init__(provider)
        if provider != "openai":
            raise ValueError(
                "OpenAIEmbeddingProvider must be initialized with provider `openai`."
            )
        if not os.getenv("OPENAI_API_KEY"):
            raise ValueError(
                "Must set OPENAI_API_KEY in order to initialize OpenAIEmbeddingProvider."
            )
        self.client = OpenAI()

    def _check_inputs(self, model: str, dimensions: Optional[int]) -> None:
        if model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(f"OpenAI embedding model {model} not supported.")
        if (
            dimensions
            and dimensions
            not in OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[model]
        ):
            raise ValueError(
                f"Dimensions {dimensions} for {model} are not supported"
            )
        # TODO - Check max length of tokenized documents?

    def get_embedding(
        self, text: str, model: str, dimensions: Optional[int] = None
    ) -> list[float]:
        self._check_inputs(model, dimensions)
        return (
            self.client.embeddings.create(
                input=[text],
                model=model,
                dimensions=dimensions
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[model][-1],
            )
            .data[0]
            .embedding
        )

    def get_embeddings(
        self, texts: list[str], model: str, dimensions: Optional[int] = None
    ) -> list[list[float]]:
        self._check_inputs(model, dimensions)
        return [
            ele.embedding
            for ele in self.client.embeddings.create(
                input=texts,
                model=model,
                dimensions=dimensions
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[model][-1],
            ).data
        ]

    def tokenize_string(self, text: str, model: str) -> list[int]:
        try:
            import tiktoken
        except ImportError:
            raise ValueError(
                "Must download tiktoken library to run `tokenize_string`."
            )
        # tiktoken encoding -
        # cl100k_base -	gpt-4, gpt-3.5-turbo, text-embedding-ada-002, text-embedding-3-small, text-embedding-3-large
        if model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(f"OpenAI embedding model {model} not supported.")
        encoding = tiktoken.get_encoding(
            OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER[model]
        )
        return encoding.encode(text)



from .basic.embedding import BasicDocument, BasicEmbeddingPipeline
from .basic.eval import BasicEvalPipeline
from .basic.ingestion import BasicIngestionPipeline, IngestionType
from .basic.prompt_provider import BasicPromptProvider
from .basic.rag import BasicRAGPipeline
from .web_search.rag import WebSearchRAGPipeline

__all__ = [
    "BasicDocument",
    "BasicEmbeddingPipeline",
    "BasicEvalPipeline",
    "IngestionType",
    "BasicIngestionPipeline",
    "BasicPromptProvider",
    "BasicRAGPipeline",
    "WebSearchRAGPipeline",
]



import uvicorn

from r2r.core.adapters import ReductoAdapter
from r2r.main import E2EPipelineFactory, R2RConfig
from r2r.pipelines import IngestionType

# Read more about the configuration in the documentation [https://r2r-docs.sciphi.ai/core-features/factory]
app = E2EPipelineFactory.create_pipeline(
    config=R2RConfig.load_config(),
    adapters={
        IngestionType.PDF: ReductoAdapter(),
    },
)


if __name__ == "__main__":
    # Run the FastAPI application using Uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



import logging
from typing import Optional

import modal

from r2r.core import EmbeddingProvider

logger = logging.getLogger(__name__)


class ModalEmbeddingProvider(EmbeddingProvider):
    def __init__(
        self,
        modal_app_name: str,
        modal_class_name: str,
        dimension: int,
        batch: int,
        provider: str = "modal",
    ):
        logger.info(
            "Initializing `ModalEmbeddingProvider` to provide embeddings."
        )
        super().__init__(provider)
        try:
            cls = modal.Cls.lookup(modal_app_name, modal_class_name)
            model = cls()
        except ImportError:
            raise ValueError("Unable to get modal's model cls")
        self.model = model
        self.dimension = dimension
        self.batch = batch

    def _check_inputs(self, model: str, dimensions: Optional[int]) -> None:
        if dimensions and dimensions != self.dimension:
            raise ValueError(
                f"Dimensions {dimensions} for {model} are not supported"
            )

    def get_embedding(
        self, text: str, model: str, dimensions: Optional[int] = None
    ) -> list[float]:
        raise ValueError(
            "ModalEmbeddingProvider does not support `get_embedding`."
        )

    def get_embeddings(
        self, texts: list[str], model: str, dimensions: Optional[int] = None
    ) -> list[list[float]]:
        self._check_inputs(model, dimensions)
        return self.model.completion_stream.remote(texts)

    def tokenize_string(self, text: str, model: str) -> list[int]:
        raise ValueError(
            "ModalEmbeddingProvider does not support `tokenize_string`."
        )



"""
Abstract base class for embedding pipelines.
"""

import logging
from abc import abstractmethod
from typing import Any, Optional

from ..providers.embedding import EmbeddingProvider
from ..providers.logging import LoggingDatabaseConnection
from ..providers.vector_db import VectorDBProvider, VectorEntry
from ..utils import generate_run_id
from .pipeline import Pipeline

logger = logging.getLogger(__name__)


class EmbeddingPipeline(Pipeline):
    def __init__(
        self,
        embedding_model: str,
        embeddings_provider: EmbeddingProvider,
        db: VectorDBProvider,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs,
    ):
        self.embedding_model = embedding_model
        self.embeddings_provider = embeddings_provider
        self.db = db
        super().__init__(logging_connection=logging_connection, **kwargs)

    def initialize_pipeline(self, *args, **kwargs) -> None:
        self.pipeline_run_info = {
            "run_id": generate_run_id(),
            "type": "embedding",
        }

    @abstractmethod
    def transform_text(self, text: str) -> str:
        pass

    @abstractmethod
    def chunk_text(self, text: str) -> list[str]:
        pass

    @abstractmethod
    def transform_chunks(
        self, chunks: list[Any], metadatas: list[dict]
    ) -> list[Any]:
        pass

    @abstractmethod
    def embed_chunks(self, chunks: list[Any]) -> list[list[float]]:
        pass

    @abstractmethod
    def store_chunks(self, chunks: list[VectorEntry], *args, **kwargs) -> None:
        pass

    def run(self, document: Any, **kwargs):
        self.initialize_pipeline()
        logger.debug(
            f"Running the `BasicEmbeddingPipeline` with pipeline_run_info={self.pipeline_run_info}."
        )

        documents = [document] if not isinstance(document, list) else document

        for document in documents:
            transformed_text = self.transform_text(document.text)
            chunks = self.chunk_text(transformed_text)
            transformed_chunks = self.transform_chunks(chunks, [])
            embeddings = self.embed_chunks(transformed_chunks)
            self.store_chunks(
                [
                    VectorEntry(document.id, embedding, document.metadata)
                    for embedding in embeddings
                ],
                **kwargs,
            )



from typing import Any, Optional

from r2r.core import DefaultPromptProvider


class BasicPromptProvider(DefaultPromptProvider):
    BASIC_SYSTEM_PROMPT = "You are a helpful assistant."
    BASIC_TASK_PROMPT = """
    ## Task:
    Answer the query given immediately below given the context which follows later.

    ### Query:
    {query}

    ### Context:
    {context}

    ### Query:
    {query}

    ## Response:
    """

    def __init__(
        self,
        system_prompt: Optional[str] = None,
        task_prompt: Optional[str] = None,
    ):
        super().__init__()
        self.add_prompt(
            "system_prompt", system_prompt or self.BASIC_SYSTEM_PROMPT
        )
        self.add_prompt("task_prompt", task_prompt or self.BASIC_TASK_PROMPT)

    def get_system_prompt(self, inputs: dict[str, Any]) -> str:
        return self.get_prompt("system_prompt", inputs)

    def get_task_prompt(self, inputs: dict[str, Any]) -> str:
        return self.get_prompt("task_prompt", inputs)



import functools
import json
import logging
import os
import types
from abc import ABC, abstractmethod
from datetime import datetime

logger = logging.getLogger(__name__)


class LoggingProvider(ABC):
    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def log(
        self,
        timestamp,
        pipeline_run_id,
        pipeline_run_type,
        method,
        result,
        log_level,
    ):
        pass

    @abstractmethod
    def get_logs(self, max_logs: int) -> list:
        pass


class PostgresLoggingProvider(LoggingProvider):
    def __init__(self, collection_name="logs"):
        self.conn = None
        self.collection_name = collection_name
        self.db_module = self._import_db_module()
        self._init()

    def _import_db_module(self):
        try:
            import psycopg2

            return psycopg2
        except ImportError:
            raise ValueError(
                "Error, `psycopg2` is not installed. Please install it using `pip install psycopg2`."
            )

    def _init(self):
        if not all(
            [
                os.getenv("POSTGRES_DBNAME"),
                os.getenv("POSTGRES_USER"),
                os.getenv("POSTGRES_PASSWORD"),
                os.getenv("POSTGRES_HOST"),
                os.getenv("POSTGRES_PORT"),
            ]
        ):
            raise ValueError(
                "Please set the environment variables POSTGRES_DBNAME, POSTGRES_USER, POSTGRES_PASSWORD, POSTGRES_HOST, and POSTGRES_PORT to run `LoggingDatabaseConnection` with `postgres`."
            )
        self.conn = self.db_module.connect(
            dbname=os.getenv("POSTGRES_DBNAME"),
            user=os.getenv("POSTGRES_USER"),
            password=os.getenv("POSTGRES_PASSWORD"),
            host=os.getenv("POSTGRES_HOST"),
            port=os.getenv("POSTGRES_PORT"),
        )
        with self.conn.cursor() as cur:
            cur.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self.collection_name} (
                    timestamp TIMESTAMP,
                    pipeline_run_id UUID,
                    pipeline_run_type TEXT,
                    method TEXT,
                    result TEXT,
                    log_level TEXT
                )
            """
            )
        self.conn.commit()

    def close(self):
        if self.conn:
            self.conn.close()

    def log(
        self,
        pipeline_run_id,
        pipeline_run_type,
        method,
        result,
        log_level,
    ):
        try:
            with self.conn.cursor() as cur:
                cur.execute(
                    f"INSERT INTO {self.collection_name} (timestamp, pipeline_run_id, pipeline_run_type, method, result, log_level) VALUES (NOW(), %s, %s, %s, %s, %s)",
                    (
                        str(pipeline_run_id),
                        pipeline_run_type,
                        method,
                        str(result),
                        log_level,
                    ),
                )
            self.conn.commit()
        except Exception as e:
            # Handle any exceptions that occur during the logging process
            logger.error(
                f"Error occurred while logging to the PostgreSQL database: {str(e)}"
            )

    def get_logs(self, max_logs: int) -> list:
        logs = []
        with self.db_module.connect(
            dbname=os.getenv("POSTGRES_DBNAME"),
            user=os.getenv("POSTGRES_USER"),
            password=os.getenv("POSTGRES_PASSWORD"),
            host=os.getenv("POSTGRES_HOST"),
            port=os.getenv("POSTGRES_PORT"),
        ) as conn:
            with conn.cursor() as cur:
                cur.execute(
                    f"SELECT * FROM {self.collection_name} ORDER BY timestamp DESC LIMIT %s",
                    (max_logs,),
                )
                colnames = [desc[0] for desc in cur.description]
                logs = [dict(zip(colnames, row)) for row in cur.fetchall()]
        return logs


class LocalLoggingProvider(LoggingProvider):
    def __init__(self, collection_name="logs"):
        self.conn = None
        self.collection_name = collection_name
        logging_path = os.getenv("LOCAL_DB_PATH")
        if not logging_path:
            raise ValueError(
                "Please set the environment variable LOCAL_DB_PATH to run `LoggingDatabaseConnection` with `local`."
            )
        self.logging_path = logging_path
        self.db_module = self._import_db_module()
        self._init()

    def _import_db_module(self):
        try:
            import sqlite3

            return sqlite3
        except ImportError:
            raise ValueError(
                "Error, `sqlite3` is not installed. Please install it using `pip install sqlite3`."
            )

    def _init(self):
        self.conn = self.db_module.connect(self.logging_path)
        self.conn.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.collection_name} (
                timestamp DATETIME,
                pipeline_run_id TEXT,
                pipeline_run_type TEXT,
                method TEXT,
                result TEXT,
                log_level TEXT
            )
            """
        )
        self.conn.commit()

    def close(self):
        if self.conn:
            self.conn.close()

    def log(
        self,
        pipeline_run_id,
        pipeline_run_type,
        method,
        result,
        log_level,
    ):
        try:
            with self.db_module.connect(self.logging_path) as conn:
                conn.execute(
                    f"INSERT INTO {self.collection_name} (timestamp, pipeline_run_id, pipeline_run_type, method, result, log_level) VALUES (datetime('now'), ?, ?, ?, ?, ?)",
                    (
                        str(pipeline_run_id),
                        pipeline_run_type,
                        method,
                        str(result),
                        log_level,
                    ),
                )
                conn.commit()
        except Exception as e:
            # Handle any exceptions that occur during the logging process
            logger.error(
                f"Error occurred while logging to the local database: {str(e)}"
            )

    def get_logs(self, max_logs: int) -> list:
        logs = []
        with self.db_module.connect(self.logging_path) as conn:
            cur = conn.execute(
                f"SELECT * FROM {self.collection_name} ORDER BY timestamp DESC LIMIT ?",
                (max_logs,),
            )
            colnames = [desc[0] for desc in cur.description]
            results = cur.fetchall()
            logs = [dict(zip(colnames, row)) for row in results]
        return logs


class RedisLoggingProvider(LoggingProvider):
    def __init__(self, collection_name="logs"):
        if not all(
            [
                os.getenv("REDIS_CLUSTER_IP"),
                os.getenv("REDIS_CLUSTER_PORT"),
            ]
        ):
            raise ValueError(
                "Please set the environment variables REDIS_CLUSTER_IP and REDIS_CLUSTER_PORT to run `LoggingDatabaseConnection` with `redis`."
            )
        try:
            from redis import Redis
        except ImportError:
            raise ValueError(
                "Error, `redis` is not installed. Please install it using `pip install redis`."
            )

        cluster_ip = os.getenv("REDIS_CLUSTER_IP")
        port = os.getenv("REDIS_CLUSTER_PORT")
        self.redis = Redis(cluster_ip, port, decode_responses=True)
        self.log_key = collection_name

    def connect(self):
        pass

    def close(self):
        pass

    def log(
        self,
        pipeline_run_id,
        pipeline_run_type,
        method,
        result,
        log_level,
    ):
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        log_entry = {
            "timestamp": timestamp,
            "pipeline_run_id": str(pipeline_run_id),
            "pipeline_run_type": pipeline_run_type,
            "method": method,
            "result": str(result),
            "log_level": log_level,
        }
        try:
            self.redis.lpush(self.log_key, json.dumps(log_entry))
        except Exception as e:
            # Handle any exceptions that occur during the logging process
            logger.error(f"Error occurred while logging to Redis: {str(e)}")

    def get_logs(self, max_logs: int) -> list:
        logs = self.redis.lrange(self.log_key, 0, max_logs - 1)
        return [json.loads(log) for log in logs]


class LoggingDatabaseConnection:
    """
    A class to connect to a database and log the execution of methods to it.
    """

    supported_providers = {
        "postgres": PostgresLoggingProvider,
        "local": LocalLoggingProvider,
        "redis": RedisLoggingProvider,
    }

    def __init__(
        self,
        provider="postgres",
        collection_name="logs",
    ):
        if provider not in self.supported_providers:
            raise ValueError(
                f"Error, `{provider}` is not in LoggingDatabaseConnection's list of supported providers."
            )

        self.provider = provider
        if provider == "redis":
            self.logging_provider = self.supported_providers[provider](
                collection_name
            )
        else:
            self.logging_provider = self.supported_providers[provider](
                collection_name
            )

    def __enter__(self):
        self.logging_provider.connect()
        return self.logging_provider

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logging_provider.close()

    def get_logs(self, max_logs: int) -> list:
        return self.logging_provider.get_logs(max_logs)


def log_execution_to_db(func):
    """A decorator to log the execution of a method to the database."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        instance = args[0]
        logging_connection = instance.logging_connection
        if not logging_connection:
            return func(*args, **kwargs)

        arg_pipeline_run_id = instance.pipeline_run_info["run_id"]
        arg_pipeline_run_type = instance.pipeline_run_info["type"]

        try:
            result = func(*args, **kwargs)
            if isinstance(result, types.GeneratorType):

                def generator_wrapper():
                    log_level = "INFO"
                    results = []
                    try:
                        for res in result:
                            results.append(res)
                            yield res
                    except Exception as e:
                        results.append(str(e))
                        log_level = "ERROR"
                    finally:
                        logging_connection.logging_provider.log(
                            arg_pipeline_run_id,
                            arg_pipeline_run_type,
                            func.__name__,
                            "".join(results),
                            log_level,
                        )

                return generator_wrapper()
            else:
                logging_connection.logging_provider.log(
                    arg_pipeline_run_id,
                    arg_pipeline_run_type,
                    func.__name__,
                    str(result),
                    "INFO",
                )
                return result

        except Exception as e:
            result = str(e)
            log_level = "ERROR"
            logging_connection.logging_provider.log(
                arg_pipeline_run_id,
                arg_pipeline_run_type,
                func.__name__,
                result,
                log_level,
            )
            raise Exception(result)

    return wrapper



"""Base classes for language model providers."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from typing import Optional

from openai.types.chat import ChatCompletion, ChatCompletionChunk


@dataclass
class LLMConfig(ABC):
    provider: Optional[str] = None
    version: str = "0.1.0"

    @classmethod
    def create(cls, **kwargs):
        valid_keys = {f.name for f in fields(cls)}
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in valid_keys}
        return cls(**filtered_kwargs)


@dataclass
class GenerationConfig(ABC):
    temperature: float = 0.1
    top_p: float = 1.0
    top_k: int = 100
    max_tokens_to_sample: int = 1_024
    model: Optional[str] = None
    stream: bool = False
    functions: Optional[list[dict]] = None
    skip_special_tokens: bool = False
    stop_token: Optional[str] = None
    num_beams: int = 1
    do_sample: bool = True
    # Additional args to pass to the generation config
    add_generation_kwargs: dict = field(default_factory=dict)
    generate_with_chat: bool = False
    api_base: Optional[str] = None


class LLMProvider(ABC):
    """An abstract class to provide a common interface for LLMs."""

    def __init__(
        self,
    ) -> None:
        pass

    @abstractmethod
    def get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletion:
        """Abstract method to get a chat completion from the provider."""
        pass

    @abstractmethod
    def get_completion_stream(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletionChunk:
        """Abstract method to get a completion stream from the provider."""
        pass



from abc import ABC, abstractmethod
from typing import Optional


class EmbeddingProvider(ABC):
    supported_providers = ["openai", "sentence-transformers"]

    def __init__(self, provider: str):
        if provider not in EmbeddingProvider.supported_providers:
            raise ValueError(
                f"Error, `{provider}` is not in EmbeddingProvider's list of supported providers."
            )

    @abstractmethod
    def _check_inputs(self, model: str, dimensions: Optional[int]) -> None:
        pass

    @abstractmethod
    def get_embedding(
        self, text: str, model: str, dimensions: Optional[int] = None
    ):
        pass

    @abstractmethod
    def get_embeddings(
        self, texts: list[str], model: str, dimensions: Optional[int] = None
    ):
        pass

    @abstractmethod
    def tokenize_string(self, text: str, model: str) -> list[int]:
        """Tokenizes the input string."""
        pass



import json
from typing import Any, Dict, List, Optional, Union

import httpx
import requests


class R2RClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def upload_and_process_file(
        self,
        document_id: str,
        file_path: str,
        metadata: Optional[dict] = None,
        settings: Optional[dict] = None,
    ):
        url = f"{self.base_url}/upload_and_process_file/"
        with open(file_path, "rb") as file:
            files = {
                "file": (file_path.split("/")[-1], file, "application/pdf")
            }
            data = {
                "document_id": document_id,
                "metadata": (
                    json.dumps(metadata) if metadata else json.dumps({})
                ),
                "settings": (
                    json.dumps(settings) if settings else json.dumps({})
                ),
            }
            response = requests.post(
                url,
                files=files,
                data=data,
            )
        return response.json()

    def add_entry(
        self,
        document_id: str,
        blobs: Dict[str, str],
        metadata: Optional[Dict[str, Any]] = None,
        do_upsert: Optional[bool] = False,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/add_entry/"
        json_data = {
            "entry": {
                "document_id": document_id,
                "blobs": blobs,
                "metadata": metadata or {},
            },
            "settings": settings
            or {"embedding_settings": {"do_upsert": do_upsert}},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def add_entries(
        self,
        entries: List[Dict[str, Any]],
        do_upsert: Optional[bool] = False,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/add_entries/"
        json_data = {
            "entries": entries,
            "settings": settings
            or {"embedding_settings": {"do_upsert": do_upsert}},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def search(
        self,
        query: str,
        limit: Optional[int] = 10,
        filters: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/search/"
        json_data = {
            "query": query,
            "filters": filters or {},
            "limit": limit,
            "settings": settings or {},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    # TODO - Cleanup redundant code in the following methods
    # TODO - Consider how to improve `rag_completion` and
    # `stream_rag_completion` workflows.
    def rag_completion(
        self,
        query: str,
        limit: Optional[int] = 10,
        filters: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
    ):
        if not generation_config:
            generation_config = {}
        stream = generation_config.get("stream", False)

        if stream:
            raise ValueError(
                "To stream, use the `stream_rag_completion` method."
            )

        url = f"{self.base_url}/rag_completion/"
        json_data = {
            "query": query,
            "filters": filters or {},
            "limit": limit,
            "settings": settings or {},
            "generation_config": generation_config or {},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def eval(
        self,
        query: str,
        context: str,
        completion_text: str,
        run_id: str,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/eval/"
        payload = {
            "query": query,
            "context": context,
            "completion_text": completion_text,
            "run_id": run_id,
            "settings": settings or {},
        }
        response = requests.post(url, json=payload)
        return response.json()

    async def stream_rag_completion(
        self,
        query: str,
        limit: Optional[int] = 10,
        filters: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
    ):
        if not generation_config:
            generation_config = {}
        stream = generation_config.get("stream", False)
        if not stream:
            raise ValueError(
                "`stream_rag_completion` method is only for streaming."
            )

        url = f"{self.base_url}/rag_completion/"

        headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
        }
        json_data = {
            "query": query,
            "filters": filters or {},
            "limit": limit,
            "settings": settings or {},
            "generation_config": generation_config or {},
        }

        async with httpx.AsyncClient() as client:
            async with client.stream(
                "POST", url, headers=headers, json=json_data
            ) as response:
                async for chunk in response.aiter_bytes():
                    yield chunk.decode()

    def filtered_deletion(self, key: str, value: Union[bool, int, str]):
        url = f"{self.base_url}/filtered_deletion/"
        response = requests.delete(url, params={"key": key, "value": value})
        return response.json()

    def get_logs(self):
        url = f"{self.base_url}/logs"
        response = requests.get(url)
        return response.json()

    def get_logs_summary(self):
        url = f"{self.base_url}/logs_summary"
        response = requests.get(url)
        return response.json()



from abc import ABC, abstractmethod
from typing import Optional

from ..providers.logging import LoggingDatabaseConnection


class Pipeline(ABC):
    def __init__(
        self,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs
    ):
        self.logging_connection = logging_connection
        self.pipeline_run_info: Optional[dict] = None

    def _check_pipeline_initialized(self) -> None:
        if self.pipeline_run_info is None:
            raise ValueError(
                "The pipeline has not been initialized. Please call `initialize_pipeline` before running the pipeline."
            )

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)
        if isinstance(attr, property):
            return attr

        if callable(attr) and name not in [
            "__init__",
            "close",
            "_check_pipeline_initialized",
            "__getattribute__",
            "initialize_pipeline",
            "run",
        ]:

            def newfunc(*args, **kwargs):
                self._check_pipeline_initialized()
                return attr(*args, **kwargs)

            return newfunc
        else:
            return attr

    def close(self):
        if self.logging_connection:
            self.logging_connection.__exit__(None, None, None)

    @abstractmethod
    def initialize_pipeline(self, *args, **kwargs):
        pass

    @abstractmethod
    def run(self, *args, **kwargs):
        pass



import json
import logging
import os
import re
from logging.handlers import RotatingFileHandler
from typing import Any, Union

from fastapi.middleware.cors import CORSMiddleware

from r2r.core import BasicDocument

logger = logging.getLogger(__name__)


# Function to find the project root by looking for a .git folder or setup.py file
def find_project_root(current_dir):
    for parent in current_dir.parents:
        if any((parent / marker).exists() for marker in [".git", "setup.py"]):
            return parent
    return current_dir  # Fallback to current dir if no marker found


REQUIRED_KEYS: dict[str, list] = {
    "embedding": ["provider", "model"],
    "evals": ["provider", "frequency"],
    "language_model": ["provider"],
    "logging_database": ["provider", "collection_name", "level"],
    "ingestion": ["provider"],
    "vector_database": ["provider", "collection_name"],
    "app": ["max_logs"],
}


class R2RConfig:
    def __init__(self, config_data: dict[str, Any]):
        for section, keys in REQUIRED_KEYS.items():
            self._validate_config_section(config_data, section, keys)
            setattr(self, section, config_data[section])

    def _validate_config_section(
        self, config_data: dict[str, Any], section: str, keys: list
    ):
        if section not in config_data:
            raise ValueError(f"Missing '{section}' section in config")
        if not all(key in config_data[section] for key in keys):
            raise ValueError(f"Missing required keys in '{section}' config")

    @classmethod
    def load_config(cls, config_path: str = None) -> "R2RConfig":
        if config_path is None:
            # Get the root directory of the project
            root_dir = os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            config_path = os.path.join(root_dir, "config.json")

        # Load configuration from JSON file
        with open(config_path) as f:
            config_data = json.load(f)

        return cls(config_data)

    # TODO - How to type 'redis.Redis' without introducing dependency on 'redis' package?
    def save_to_redis(self, redis_client: Any, key: str):
        config_data = {
            section: getattr(self, section) for section in REQUIRED_KEYS.keys()
        }
        redis_client.set(f"R2RConfig:{key}", json.dumps(config_data))

    @classmethod
    def load_from_redis(cls, redis_client: Any, key: str) -> "R2RConfig":
        config_data = redis_client.get(f"R2RConfig:{key}")
        if config_data is None:
            raise ValueError(
                f"Configuration not found in Redis with key '{key}'"
            )
        return cls(json.loads(config_data))


def apply_cors(app):
    # CORS setup
    origins = [
        "*",  # TODO - Change this to the actual frontend URL
        "http://localhost:3000",
        "http://localhost:8000",
    ]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,  # Allows specified origins
        allow_credentials=True,
        allow_methods=["*"],  # Allows all methods
        allow_headers=["*"],  # Allows all headers
    )


def configure_logging():
    # Determine the root directory of the project
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    logs_dir = os.path.join(
        root_dir, "..", "logs"
    )  # Place the logs directory one level above the root directory

    # Create the logs directory if it doesn't exist
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    # Create a custom logger
    logger = logging.getLogger("r2r")
    logger.setLevel(logging.DEBUG)  # Set the logging level

    # Create handlers (console and file handler with rotation)
    c_handler = logging.StreamHandler()
    log_file_path = os.path.join(logs_dir, "r2r.log")
    f_handler = RotatingFileHandler(
        log_file_path, maxBytes=1000000, backupCount=5
    )
    c_handler.setLevel(logging.WARNING)  # Console handler level
    f_handler.setLevel(logging.DEBUG)  # File handler level

    # Create formatters and add it to handlers
    c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    f_format = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)

    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)


def update_aggregation_entries(
    log: dict[str, Any],
    event_aggregation: dict[str, dict[str, Union[str, list]]],
):
    pipeline_run_id = log["pipeline_run_id"]
    if pipeline_run_id is None:
        logger.error(f"Missing 'pipeline_run_id' in log: {log}")
        raise ValueError(f"Missing 'pipeline_run_id' in log: {log}")

    pipeline_run_type = log["pipeline_run_type"]
    if pipeline_run_type is None:
        logger.error(f"Missing 'pipeline_run_type' in log: {log}")
        raise ValueError(f"Missing 'pipeline_run_type' in log: {log}")

    if pipeline_run_id not in event_aggregation:
        event_aggregation[pipeline_run_id] = {
            "timestamp": log["timestamp"],
            "pipeline_run_id": pipeline_run_id,
            "pipeline_run_type": pipeline_run_type,
            "events": [],
        }
    event = {
        "method": log["method"],
        "result": log["result"],
        "log_level": log["log_level"],
        "outcome": "success" if log["log_level"] == "INFO" else "fail",
    }
    if (
        pipeline_run_id not in event_aggregation
        or "events" not in event_aggregation[pipeline_run_id]
    ):
        if isinstance(event_aggregation[pipeline_run_id]["events"], list):
            raise ValueError(f"Incorrect 'events' datatype event_aggregation")

        raise ValueError(
            f"Missing 'pipeline_run_id' in event_aggregation: {event_aggregation}"
        )

    event_aggregation[pipeline_run_id]["events"].append(event)  # type: ignore


def process_event(event: dict[str, Any], pipeline_type: str) -> dict[str, Any]:
    method = event["method"]
    result = event.get("result", "N/A")
    processed_result = {}

    if method == "ingress":
        try:
            processed_result["search_query"] = result
        except Exception as e:
            logger.error(f"Error {e} processing 'ingress' event: {event}")
    elif method == "ingress" and pipeline_type == "embedding":
        try:
            id_match = re.search(r"'id': '([^']+)'", result)
            text_match = re.search(r"'text': '([^']+)'", result)
            metadata_match = re.search(r"'metadata': (\{[^}]+\})", result)
            if not id_match or not text_match or not metadata_match:
                raise ValueError(
                    f"Missing 'id', 'text', or 'metadata' in result: {result}"
                )
            metadata = metadata_match.group(1).replace("'", '"')
            metadata_json = json.loads(metadata)

            processed_result["document"] = BasicDocument(
                id=id_match.group(1),
                text=text_match.group(1),
                metadata=metadata_json,
            )
        except Exception as e:
            logger.error(f"Error {e} processing 'ingress' event: {event}")
    elif method == "search":
        try:
            text_matches = re.findall(r"'text': '([^']*)'", result)
            scores = re.findall(r"score=(\d+\.\d+)", result)
            processed_result["search_results"] = [
                {"text": text, "score": score}
                for text, score in zip(text_matches, scores)
            ]
            processed_result["method"] = "Search"
        except Exception as e:
            logger.error(f"Error {e} processing 'ingress' event: {event}")

    elif method == "generate_completion":
        try:
            if "content=" in result:
                content_matches = re.findall(r'content="([^"]*)"', result)
                processed_result["completion_result"] = ", ".join(
                    content_matches
                )
            else:
                processed_result["completion_result"] = result
            processed_result["method"] = "RAG"
        except Exception as e:
            logger.error(
                f"Error {e} processing 'generate_completion' event: {event}"
            )
    elif method == "evaluate":
        try:
            result = result.replace("'", '"')  # Convert to valid JSON string
            processed_result["eval_results"] = json.loads(result)
        except Exception as e:
            logger.error(f"Error {e} decoding JSON: {e}")
    elif method == "transform_chunks":
        try:
            processed_result["method"] = "Embedding"
            processed_result["embedding_chunks"] = result
        except Exception as e:
            logger.error(
                f"Error {e} processing 'transform_chunks' event: {event}"
            )

    return processed_result


def combine_aggregated_logs(
    event_aggregation: dict[str, dict[str, Any]]
) -> list[dict[str, Any]]:
    logs_summary = []
    for run_id, aggregation in event_aggregation.items():
        # Assuming 'pipeline_run_type' is available in the log entries to determine the type of pipeline
        pipeline_type = (
            aggregation["pipeline_run_type"]
            if "pipeline_run_type" in aggregation
            else "unknown"
        )

        summary_entry = {
            "timestamp": aggregation["timestamp"],
            "pipeline_run_id": run_id,
            "pipeline_run_type": pipeline_type,
            "method": "",
            "search_query": "",
            "search_results": [],
            "eval_results": None,
            "embedding_chunks": None,
            "document": None,
            "completion_result": "N/A",
            "outcome": (
                "success"
                if aggregation["events"][-1].get("log_level") == "INFO"
                else "fail"
            ),
        }

        for event in aggregation["events"]:
            new_event = process_event(event, pipeline_type)
            if summary_entry["embedding_chunks"]:
                new_event["embedding_chunks"] = summary_entry[
                    "embedding_chunks"
                ] + new_event.get("embedding_chunks", "")

            if summary_entry["document"] is not None:
                new_event.pop("document", None)
            summary_entry.update(new_event)
        logs_summary.append(summary_entry)
    return logs_summary


def process_logs(logs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    event_aggregation: dict = {}
    for log in logs:
        update_aggregation_entries(log, event_aggregation)
    # Convert each aggregated log entry to SummaryLogModel before returning
    return combine_aggregated_logs(event_aggregation)



import json
import logging
from pathlib import Path
from typing import AsyncGenerator, Generator, Optional, Union, cast

import requests
from fastapi import (
    BackgroundTasks,
    FastAPI,
    File,
    Form,
    HTTPException,
    Request,
    UploadFile,
)
from fastapi.responses import StreamingResponse

from r2r.core import (
    EmbeddingPipeline,
    EvalPipeline,
    GenerationConfig,
    IngestionPipeline,
    LoggingDatabaseConnection,
    RAGPipeline,
    RAGPipelineOutput,
)
from r2r.main.utils import (  # configure_logging,
    R2RConfig,
    apply_cors,
    find_project_root,
    process_logs,
)

from .models import (
    AddEntriesRequest,
    AddEntryRequest,
    EvalPayloadModel,
    LogModel,
    RAGQueryModel,
    SettingsModel,
    SummaryLogModel,
)

# logger = logging.getLogger("r2r")
# logging.setLevel(logging.INFO)

# Current directory where this script is located
CURRENT_DIR = Path(__file__).resolve().parent
MB_CONVERSION_FACTOR = 1024 * 1024


def create_app(
    ingestion_pipeline: IngestionPipeline,
    embedding_pipeline: EmbeddingPipeline,
    eval_pipeline: EvalPipeline,
    rag_pipeline: RAGPipeline,
    config: R2RConfig,
    upload_path: Optional[Path] = None,
    logging_connection: Optional[LoggingDatabaseConnection] = None,
):
    app = FastAPI()
    # TODO - Consider impact of logging in remote environments
    # e.g. such as Google Cloud Run
    # configure_logging()
    apply_cors(app)

    upload_path = upload_path or find_project_root(CURRENT_DIR) / "uploads"

    if not upload_path.exists():
        upload_path.mkdir()

    @app.post("/upload_and_process_file/")
    # TODO - Why can't we use a BaseModel to represent the request?
    # Naive class FileUploadRequest(BaseModel) above fails
    async def upload_and_process_file(
        document_id: str = Form(...),
        metadata: str = Form("{}"),
        settings: str = Form("{}"),
        file: UploadFile = File(...),
    ):
        metadata_json = json.loads(metadata)
        settings_model = SettingsModel.parse_raw(settings)

        if (
            file is not None
            and file.size
            > config.app.get("max_file_size_in_mb", 100) * MB_CONVERSION_FACTOR
        ):
            raise HTTPException(
                status_code=413,
                detail="File size exceeds maximum allowed size.",
            )

        if not file.filename:
            raise HTTPException(
                status_code=400, detail="No file was uploaded."
            )
        # Extract file extension and check if it's an allowed type
        file_extension = file.filename.split(".")[-1]
        if file_extension not in ingestion_pipeline.supported_types:
            raise HTTPException(
                status_code=400,
                detail=f"Invalid file type. Allowed types are: {', '.join(ingestion_pipeline.supported_types)}.",
            )

        file_location = upload_path / file.filename
        try:
            file_content = file.file.read()

            # TODO - Consider saving file to disk
            # with open(file_location, "wb+") as file_object:
            # file_object.write(file_content)

            # TODO - Mark upload as failed if ingestion or embedding fail partway through

            documents = ingestion_pipeline.run(
                document_id,
                {file_extension: file_content},
                metadata=metadata_json,
                **settings_model.ingestion_settings.dict(),
            )
            for document in documents:
                embedding_pipeline.run(
                    document, **settings_model.embedding_settings.dict()
                )

            return {
                "message": f"File '{file.filename}' processed and saved at '{file_location}'"
            }
        except Exception as e:
            logging.error(
                f"upload_and_process_file: [Error](file={file.filename}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/add_entry/")
    async def add_entry(entry_req: AddEntryRequest):
        try:
            documents = ingestion_pipeline.run(
                entry_req.entry.document_id,
                entry_req.entry.blobs,
                metadata=entry_req.entry.metadata,
                **entry_req.settings.ingestion_settings.dict(),
            )
            for document in documents:
                embedding_pipeline.run(
                    document, **entry_req.settings.embedding_settings.dict()
                )
            return {"message": "Entry upserted successfully."}
        except Exception as e:
            logging.error(
                f":add_entry: [Error](entry={entry_req}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/add_entries/")
    async def add_entries(entries_req: AddEntriesRequest):
        try:
            for entry in entries_req.entries:
                documents = ingestion_pipeline.run(
                    entry.document_id,
                    entry.blobs,
                    metadata=entry.metadata,
                    **entries_req.settings.ingestion_settings.dict(),
                )
                for document in documents:
                    embedding_pipeline.run(
                        document,
                        **entries_req.settings.embedding_settings.dict(),
                    )
            return {"message": "Entries upserted successfully."}
        except Exception as e:
            logging.error(
                f":add_entries: [Error](entries={entries_req}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/search/")
    async def search(query: RAGQueryModel):
        try:
            rag_completion = rag_pipeline.run(
                query.query, query.filters, query.limit, search_only=True
            )
            return rag_completion.search_results
        except Exception as e:
            logging.error(f":search: [Error](query={query}, error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/rag_completion/")
    async def rag_completion(
        background_tasks: BackgroundTasks,
        query: RAGQueryModel,
        request: Request,
    ):
        try:
            stream = query.generation_config.stream
            if not stream:
                untyped_completion = rag_pipeline.run(
                    query.query,
                    query.filters,
                    query.limit,
                    generation_config=query.generation_config,
                )
                # Tell the type checker that rag_completion is a RAGPipelineOutput
                rag_completion = cast(RAGPipelineOutput, untyped_completion)
                if not rag_completion.completion:
                    raise ValueError(
                        "No completion found in RAGPipelineOutput."
                    )

                completion_text = rag_completion.completion.choices[
                    0
                ].message.content

                # Retrieve the URL dynamically from the request header
                url = request.url
                if not url:
                    url = "http://localhost:8000"
                else:
                    url = str(url).split("/rag_completion")[0]

                # Pass the payload to the /eval endpoint
                payload = {
                    "query": query.query,
                    "context": rag_completion.context or "",
                    "completion_text": completion_text or "",
                    "run_id": str(rag_pipeline.pipeline_run_info["run_id"]),
                    "settings": query.settings.rag_settings.dict(),
                }
                if config.evals.get("frequency", 0.0) > 0.0:
                    background_tasks.add_task(
                        requests.post, f"{url}/eval", json=payload
                    )

                return rag_completion

            else:

                async def _stream_rag_completion(
                    query: RAGQueryModel,
                    rag_pipeline: RAGPipeline,
                ) -> AsyncGenerator[str, None]:
                    gen_config = GenerationConfig(
                        **(query.generation_config.dict())
                    )
                    if not gen_config.stream:
                        raise ValueError(
                            "Must pass `stream` as True to stream completions."
                        )
                    completion_generator = cast(
                        Generator[str, None, None],
                        rag_pipeline.run(
                            query.query,
                            query.filters,
                            query.limit,
                            generation_config=gen_config,
                        ),
                    )

                    search_results = ""
                    context = ""
                    completion_text = ""
                    current_marker = None

                    logging.info(
                        f"Streaming RAG completion results to client for query ={query.query}."
                    )

                    for item in completion_generator:
                        if item.startswith("<"):
                            if item.startswith(
                                f"<{RAGPipeline.SEARCH_STREAM_MARKER}>"
                            ):
                                current_marker = (
                                    RAGPipeline.SEARCH_STREAM_MARKER
                                )
                            elif item.startswith(
                                f"<{RAGPipeline.CONTEXT_STREAM_MARKER}>"
                            ):
                                current_marker = (
                                    RAGPipeline.CONTEXT_STREAM_MARKER
                                )
                            elif item.startswith(
                                f"<{RAGPipeline.COMPLETION_STREAM_MARKER}>"
                            ):
                                current_marker = (
                                    RAGPipeline.COMPLETION_STREAM_MARKER
                                )
                            else:
                                current_marker = None
                        else:
                            if (
                                current_marker
                                == RAGPipeline.SEARCH_STREAM_MARKER
                            ):
                                search_results += item
                            elif (
                                current_marker
                                == RAGPipeline.CONTEXT_STREAM_MARKER
                            ):
                                context += item
                            elif (
                                current_marker
                                == RAGPipeline.COMPLETION_STREAM_MARKER
                            ):
                                completion_text += item
                        yield item

                    # Retrieve the URL dynamically from the request header
                    url = request.url
                    if not url:
                        url = "http://localhost:8000"
                    else:
                        url = str(url).split("/rag_completion")[0]
                        if "localhost" not in url and "127.0.0.1" not in url:
                            url = url.replace("http://", "https://")

                    # Pass the payload to the /eval endpoint
                    payload = {
                        "query": query.query,
                        "context": context,
                        "completion_text": completion_text,
                        "run_id": str(
                            rag_pipeline.pipeline_run_info["run_id"]
                        ),
                        "settings": query.settings.rag_settings.dict(),
                    }
                    logging.info(
                        f"Performing evaluation with payload: {payload} to url: {url}/eval"
                    )
                    background_tasks.add_task(
                        requests.post, f"{url}/eval", json=payload
                    )

                return StreamingResponse(
                    _stream_rag_completion(query, rag_pipeline),
                    media_type="text/plain",
                )
        except Exception as e:
            logging.error(
                f":rag_completion: [Error](query={query}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/eval")
    async def eval(payload: EvalPayloadModel):
        try:
            logging.info(
                f"Received evaluation payload: {payload.dict(exclude_none=True)}"
            )
            query = payload.query
            context = payload.context
            completion_text = payload.completion_text
            run_id = payload.run_id
            settings = payload.settings

            eval_pipeline.run(
                query, context, completion_text, run_id, **(settings.dict())
            )

            return {"message": "Evaluation completed successfully."}
        except Exception as e:
            logging.error(
                f":eval_endpoint: [Error](payload={payload}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.delete("/filtered_deletion/")
    async def filtered_deletion(key: str, value: Union[bool, int, str]):
        try:
            embedding_pipeline.db.filtered_deletion(key, value)
            return {"message": "Entries deleted successfully."}
        except Exception as e:
            logging.error(
                f":filtered_deletion: [Error](key={key}, value={value}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/get_user_ids/")
    async def get_user_ids():
        try:
            user_ids = embedding_pipeline.db.get_all_unique_values(
                metadata_field="user_id"
            )
            return {"user_ids": user_ids}
        except Exception as e:
            logging.error(f":get_user_ids: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/get_user_documents/")
    async def get_user_documents(user_id: str):
        try:
            document_ids = embedding_pipeline.db.get_all_unique_values(
                metadata_field="document_id", filters={"user_id": user_id}
            )
            return {"document_ids": document_ids}
        except Exception as e:
            logging.error(f":get_user_documents: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/logs")
    async def logs():
        try:
            if logging_connection is None:
                raise HTTPException(
                    status_code=404, detail="Logging provider not found."
                )
            logs = logging_connection.get_logs(config.app["max_logs"])
            for log in logs:
                LogModel(**log).dict(by_alias=True)
            return {
                "logs": [LogModel(**log).dict(by_alias=True) for log in logs]
            }
        except Exception as e:
            logging.error(f":logs: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/logs_summary")
    async def logs_summary():
        try:
            if logging_connection is None:
                raise HTTPException(
                    status_code=404, detail="Logging provider not found."
                )
            logs = logging_connection.get_logs(config.app["max_logs"])
            logs_summary = process_logs(logs)
            events_summary = [
                SummaryLogModel(**log).dict(by_alias=True)
                for log in logs_summary
            ]
            return {"events_summary": events_summary}

        except Exception as e:
            logging.error(f":logs_summary: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    return app



from .app import create_app
from .factory import E2EPipelineFactory
from .utils import R2RConfig, configure_logging

__all__ = [
    "create_app",
    "configure_logging",
    "R2RConfig",
    "E2EPipelineFactory",
]



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from r2r.core import BasicDocument


class EmbeddingsSettingsModel(BaseModel):
    do_chunking: Optional[bool] = True
    do_upsert: Optional[bool] = True


class IngestionSettingsModel(BaseModel):
    pass


class RAGSettingsModel(BaseModel):
    pass


class SettingsModel(BaseModel):
    embedding_settings: EmbeddingsSettingsModel = EmbeddingsSettingsModel()
    ingestion_settings: IngestionSettingsModel = IngestionSettingsModel()
    rag_settings: RAGSettingsModel = RAGSettingsModel()


class EntryModel(BaseModel):
    document_id: str
    blobs: dict[str, str]
    metadata: Optional[dict]


class EvalPayloadModel(BaseModel):
    query: str
    context: str
    completion_text: str
    run_id: str
    settings: RAGSettingsModel


# File upload class, to be fixed later
# class FileUploadRequest(BaseModel):
# document_id: str
# metadata: Optional[dict]
# settings: SettingsModel = SettingsModel()


class AddEntryRequest(BaseModel):
    entry: EntryModel
    settings: SettingsModel = SettingsModel()


class AddEntriesRequest(BaseModel):
    entries: list[EntryModel]
    settings: SettingsModel = SettingsModel()


class GenerationConfigModel(BaseModel):
    temperature: float = 0.1
    top_p: float = 1.0
    top_k: int = 100
    max_tokens_to_sample: int = 1_024
    model: Optional[str] = "gpt-4-0125-preview"
    stream: bool = False
    functions: Optional[list[dict]] = None
    skip_special_tokens: bool = False
    stop_token: Optional[str] = None
    num_beams: int = 1
    do_sample: bool = True
    add_generation_kwargs: dict = {}
    generate_with_chat: bool = False


class RAGQueryModel(BaseModel):
    query: str
    limit: Optional[int] = 10
    filters: dict = {}
    settings: SettingsModel = SettingsModel()
    generation_config: GenerationConfigModel = GenerationConfigModel()


def to_camel(string: str) -> str:
    return "".join(
        word.capitalize() if i != 0 else word
        for i, word in enumerate(string.split("_"))
    )


class LogModel(BaseModel):
    timestamp: datetime = Field(alias="timestamp")
    pipeline_run_id: str = Field(alias="pipelineRunId")
    pipeline_run_type: str = Field(alias="pipelineRunType")
    method: str = Field(alias="method")
    result: str = Field(alias="result")
    log_level: str = Field(alias="logLevel")

    class Config:
        alias_generator = to_camel
        populate_by_name = True


# TODO - Split apart `retrieval` and `embedding` event logs
class SummaryLogModel(BaseModel):
    timestamp: datetime = Field(alias="timestamp")
    pipeline_run_id: str = Field(alias="pipelineRunId")
    pipeline_run_type: str = Field(alias="pipelineRunType")
    method: str = Field(alias="method")
    embedding_chunks: Optional[str] = Field(alias="embeddingChunks")
    search_query: str = Field(alias="searchQuery")
    search_results: list[dict] = Field(alias="searchResults")
    completion_result: str = Field(alias="completionResult")
    eval_results: Optional[dict] = Field(alias="evalResults")
    document: Optional[BasicDocument] = Field(alias="document")
    outcome: str = Field(alias="outcome")

    class Config:
        alias_generator = to_camel
        populate_by_name = True



import http.client
import json
import os


# TODO - Move process json to dedicated data processing module
def process_json(json_object, indent=0):
    """
    Recursively traverses the JSON object (dicts and lists) to create an unstructured text blob.
    """
    text_blob = ""
    if isinstance(json_object, dict):
        for key, value in json_object.items():
            padding = "  " * indent
            if isinstance(value, (dict, list)):
                text_blob += (
                    f"{padding}{key}:\n{process_json(value, indent + 1)}"
                )
            else:
                text_blob += f"{padding}{key}: {value}\n"
    elif isinstance(json_object, list):
        for index, item in enumerate(json_object):
            padding = "  " * indent
            if isinstance(item, (dict, list)):
                text_blob += f"{padding}Item {index + 1}:\n{process_json(item, indent + 1)}"
            else:
                text_blob += f"{padding}Item {index + 1}: {item}\n"
    return text_blob


# TODO - Introduce abstract "Integration" ABC.
class SerperClient:
    def __init__(self, api_base: str = "google.serper.dev") -> None:
        api_key = os.getenv("SERPER_API_KEY")
        if not api_key:
            raise ValueError(
                "Please set the `SERPER_API_KEY` environment variable to use `SerperClient`."
            )

        self.api_base = api_base
        self.headers = {
            "X-API-KEY": api_key,
            "Content-Type": "application/json",
        }

    @staticmethod
    def _extract_results(result_data: dict) -> list:
        formatted_results = []

        for key, value in result_data.items():
            # Skip searchParameters as it's not a result entry
            if key == "searchParameters":
                continue

            # Handle 'answerBox' as a single item
            if key == "answerBox":
                value["type"] = key  # Add the type key to the dictionary
                formatted_results.append(value)
            # Handle lists of results
            elif isinstance(value, list):
                for item in value:
                    item["type"] = key  # Add the type key to the dictionary
                    formatted_results.append(item)
            # Handle 'peopleAlsoAsk' and potentially other single item formats
            elif isinstance(value, dict):
                value["type"] = key  # Add the type key to the dictionary
                formatted_results.append(value)

        return formatted_results

    # TODO - Add explicit typing for the return value
    def get_raw(self, query: str, limit: int = 10) -> list:
        connection = http.client.HTTPSConnection(self.api_base)
        payload = json.dumps({"q": query, "num": limit})
        connection.request("POST", "/search", payload, self.headers)
        response = connection.getresponse()
        data = response.read()
        json_data = json.loads(data.decode("utf-8"))
        return SerperClient._extract_results(json_data)

    @staticmethod
    def construct_context(results: list) -> str:
        # Organize results by type
        organized_results = {}
        for result in results:
            result_type = result.pop(
                "type", "Unknown"
            )  # Pop the type and use as key
            if result_type not in organized_results:
                organized_results[result_type] = [result]
            else:
                organized_results[result_type].append(result)

        context = ""
        # Iterate over each result type
        for result_type, items in organized_results.items():
            context += f"# {result_type} Results:\n"
            for index, item in enumerate(items, start=1):
                # Process each item under the current type
                context += f"Item {index}:\n"
                context += process_json(item) + "\n"

        return context



"""
A simple example to demonstrate the usage of `BasicEmbeddingPipeline`.
"""

import copy
import logging
from typing import Any, Optional, Tuple

from r2r.core import (
    BasicDocument,
    EmbeddingPipeline,
    LoggingDatabaseConnection,
    VectorDBProvider,
    VectorEntry,
    log_execution_to_db,
)
from r2r.core.utils import TextSplitter, generate_id_from_label
from r2r.embeddings import OpenAIEmbeddingProvider

logger = logging.getLogger(__name__)


class BasicEmbeddingPipeline(EmbeddingPipeline):
    """
    Embeds and stores documents using a specified embedding model and database.
    """

    def __init__(
        self,
        embedding_model: str,
        embeddings_provider: OpenAIEmbeddingProvider,
        db: VectorDBProvider,
        text_splitter: TextSplitter,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        embedding_batch_size: int = 1,
        id_prefix: str = "demo",
    ):
        """
        Initializes the embedding pipeline with necessary components and configurations.
        """
        logger.info(
            f"Initalizing a `BasicEmbeddingPipeline` to embed and store documents."
        )

        super().__init__(
            embedding_model,
            embeddings_provider,
            db,
            logging_connection,
        )
        self.text_splitter = text_splitter
        self.embedding_batch_size = embedding_batch_size
        self.id_prefix = id_prefix
        self.pipeline_run_info = None

    @log_execution_to_db
    def ingress(self, document: BasicDocument) -> dict:
        """
        Extracts text from a document.
        """
        return {
            "id": str(document.id),
            "metadata": document.metadata,
            "text": document.text,
        }

    def initialize_pipeline(
        self, document: BasicDocument, *args, **kwargs
    ) -> None:
        super().initialize_pipeline(*args, **kwargs)
        self.ingress(document)

    def transform_text(self, text: str) -> str:
        """
        Transforms text before chunking, if necessary.
        """
        return text

    def chunk_text(self, text: str) -> list[str]:
        """
        Splits text into manageable chunks for embedding.
        """
        return [
            ele.page_content
            for ele in self.text_splitter.create_documents([text])
        ]

    @log_execution_to_db
    def transform_chunks(
        self, chunks: list[str], metadatas: list[dict]
    ) -> list[str]:
        """
        Transforms text chunks based on their metadata, e.g., adding prefixes.
        """
        transformed_chunks = []
        for chunk, metadata in zip(chunks, metadatas):
            if "chunk_prefix" in metadata:
                prefix = metadata.pop("chunk_prefix")
                transformed_chunks.append(f"{prefix}\n{chunk}")
            else:
                transformed_chunks.append(chunk)
        return transformed_chunks

    def embed_chunks(self, chunks: list[str]) -> list[list[float]]:
        """
        Generates embeddings for each text chunk using the embedding model.
        """
        return self.embeddings_provider.get_embeddings(
            chunks, self.embedding_model
        )

    def store_chunks(
        self, chunks: list[VectorEntry], do_upsert: bool, *args, **kwargs
    ) -> None:
        """
        Stores the embedded chunks in the database, with an option to upsert.
        """
        if do_upsert:
            self.db.upsert_entries(chunks)
        else:
            self.db.copy_entries(chunks)

    def run(
        self,
        document: BasicDocument,
        do_chunking=False,
        do_upsert=True,
        **kwargs: Any,
    ):
        """
        Executes the embedding pipeline: chunking, transforming, embedding, and storing documents.
        """
        self.initialize_pipeline(document=document)

        logger.debug(
            f"Running the `BasicEmbeddingPipeline` with pipeline_run_info={self.pipeline_run_info}."
        )

        documents = [document] if not isinstance(document, list) else document
        batch_data = []

        for document in documents:
            chunks = (
                self.chunk_text(document.text)
                if do_chunking
                else [document.text]
            )
            for chunk in chunks:
                batch_data.append(
                    (document.id, chunk, copy.copy(document.metadata))
                )

                if len(batch_data) == self.embedding_batch_size:
                    self._process_batches(batch_data, do_upsert)
                    batch_data = []

        # Process any remaining batch
        if batch_data:
            self._process_batches(batch_data, do_upsert)

    def _process_batches(
        self, batch_data: list[Tuple[str, str, dict]], do_upsert: bool
    ):
        """
        Processes batches of documents: transforms, embeds, and stores chunks.
        """
        logger.debug(f"Parsing batch of size {len(batch_data)}.")

        entries = []

        # Unpack document IDs, indices, and chunks for transformation and embedding
        ids, raw_chunks, metadatas = zip(*batch_data)
        transformed_chunks = self.transform_chunks(raw_chunks, metadatas)
        embedded_chunks = self.embed_chunks(transformed_chunks)

        chunk_count = 0
        for doc_id, transformed_chunk, embedded_chunk, metadata in zip(
            ids, transformed_chunks, embedded_chunks, metadatas
        ):
            metadata = copy.deepcopy(metadata)
            metadata["pipeline_run_id"] = str(self.pipeline_run_info["run_id"])  # type: ignore
            metadata["text"] = transformed_chunk
            metadata["document_id"] = doc_id
            chunk_id = generate_id_from_label(f"{doc_id}-{chunk_count}")
            chunk_count += 1
            entries.append(VectorEntry(chunk_id, embedded_chunk, metadata))
        self.store_chunks(entries, do_upsert)



from .advanced.reducto import ReductoAdapter
from .base import Adapter, HTMLAdapter, JSONAdapter, PDFAdapter, TextAdapter

__all__ = [
    "Adapter",
    "PDFAdapter",
    "TextAdapter",
    "JSONAdapter",
    "HTMLAdapter",
    "ReductoAdapter",
]



import json
import logging
from typing import Generator, Optional

from r2r.core import (
    GenerationConfig,
    LLMProvider,
    LoggingDatabaseConnection,
    RAGPipeline,
    RAGPipelineOutput,
    VectorDBProvider,
    VectorSearchResult,
    log_execution_to_db,
)
from r2r.embeddings import OpenAIEmbeddingProvider
from r2r.main import E2EPipelineFactory, R2RConfig
from r2r.pipelines import BasicPromptProvider, BasicRAGPipeline

logger = logging.getLogger(__name__)

DEFAULT_SYSTEM_PROMPT = "You are a helpful assistant."
DEFAULT_TASK_PROMPT = """
## Task:
Answer the query given immediately below given the context which follows later. Use line item references to like [1], [2], ... refer to specifically numbered items in the provided context. Pay close attention to the title of each given source to ensure it is consistent with the query.

### Query:
{query}

### Context:
{context}

### Query:
{query}

REMINDER - Use line item references to like [1], [2], ... refer to specifically numbered items in the provided context.
## Response:
"""


class SyntheticRAGPipeline(BasicRAGPipeline):
    def __init__(
        self,
        llm: LLMProvider,
        db: VectorDBProvider,
        embedding_model: str,
        embeddings_provider: OpenAIEmbeddingProvider,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        system_prompt: Optional[str] = DEFAULT_SYSTEM_PROMPT,
        task_prompt: Optional[str] = DEFAULT_TASK_PROMPT,
    ) -> None:
        logger.debug(f"Initalizing `SyntheticRAGPipeline`")

        super().__init__(
            llm,
            db,
            embedding_model,
            embeddings_provider,
            logging_connection=logging_connection,
            prompt_provider=BasicPromptProvider(system_prompt, task_prompt),
        )

    def transform_query(self, query: str, generation_config: GenerationConfig) -> list[str]:  # type: ignore
        """
        Transforms the query into a list of strings.
        """
        print(f"Transforming query: {query}")
        prompt = (
            "Use the query that follows to write a comma separated three queries "
            "that will be used to retrieve the relevant materials. DO NOT generate "
            "queries which require multiple document types. E.g. ask for `lecture notes "
            "with information about X` or `readings that cover topic Y`.\n\n## Query:\n"
            "{query}\n\n## Response:".format(query=query)
        )
        orig_stream = generation_config.stream
        generation_config.stream = False
        completion = self.generate_completion(prompt, generation_config)
        transformed_queries = (
            completion.choices[0].message.content.strip().split("\n")
        )
        generation_config.stream = orig_stream
        print(f"Transformed Queries: {transformed_queries}")
        return transformed_queries

    @log_execution_to_db
    def search(
        self,
        transformed_query: str,
        filters: dict,
        limit: int,
    ) -> list[VectorSearchResult]:
        logger.debug(f"Retrieving results for query: {transformed_query}")

        results = self.db.search(
            query_vector=self.embeddings_provider.get_embedding(
                transformed_query,
                self.embedding_model,
            ),
            filters=filters,
            limit=limit,
        )
        logger.debug(f"Retrieved the raw results shown:\n{results}\n")
        return results

    @log_execution_to_db
    def construct_context(
        self,
        results: list,
    ) -> str:
        queries = [ele[0] for ele in results]
        search_results = [ele[1] for ele in results]
        reranked_results = [
            self.rerank_results(search_result)
            for search_result in search_results
        ]
        context = ""
        offset = 1
        for query, results in zip(queries, reranked_results):
            context += f"## Query:\n{query}\n\n## Context:\n{self._format_results(results, offset)}\n\n"
            offset += len(results)
        return context

    # Modifies `SyntheticRAGPipeline` run to return search_results and completion
    def run(
        self,
        query,
        filters={},
        limit=5,
        search_only=False,
        generation_config: Optional[GenerationConfig] = None,
        *args,
        **kwargs,
    ):
        """
        Runs the completion pipeline.
        """
        if not generation_config:
            generation_config = GenerationConfig(model="gpt-3.5-turbo")

        print("generation_config = ", generation_config)
        self.initialize_pipeline(query, search_only)
        transformed_queries = self.transform_query(query, generation_config)
        search_results = [
            (transformed_query, self.search(transformed_query, filters, limit))
            for transformed_query in transformed_queries
        ]
        if search_only:
            print("returning here...")
            return RAGPipelineOutput(search_results, None, None)

        context = self.construct_context(search_results)
        prompt = self.construct_prompt({"query": query, "context": context})

        print("generation_config.stream = ", generation_config.stream)
        if not generation_config.stream:
            completion = self.generate_completion(prompt, generation_config)
            print("returning there...")
            return RAGPipelineOutput(search_results, context, completion)

        return self._stream_run(
            search_results, context, prompt, generation_config
        )

    def _format_results(
        self, results: list[VectorSearchResult], start=1
    ) -> str:
        context = ""
        for i, ele in enumerate(results, start=start):
            context += f"[{i+start}] {ele.metadata['text']}\n\n"

        return context

    def _stream_run(
        self,
        search_results: list,
        context: str,
        prompt: str,
        generation_config: GenerationConfig,
    ) -> Generator[str, None, None]:
        yield f"<{RAGPipeline.SEARCH_STREAM_MARKER}>"
        yield json.dumps([ele[1][0].to_dict() for ele in search_results])
        yield f"</{RAGPipeline.SEARCH_STREAM_MARKER}>"

        yield f"<{RAGPipeline.CONTEXT_STREAM_MARKER}>"
        yield context
        yield f"</{RAGPipeline.CONTEXT_STREAM_MARKER}>"
        yield f"<{RAGPipeline.COMPLETION_STREAM_MARKER}>"
        for chunk in self.generate_completion(prompt, generation_config):
            yield chunk
        yield f"</{RAGPipeline.COMPLETION_STREAM_MARKER}>"


# Creates a pipeline using the `E2EPipelineFactory`
app = E2EPipelineFactory.create_pipeline(
    rag_pipeline_impl=SyntheticRAGPipeline,
    config=R2RConfig.load_config("config.json"),
)



from abc import abstractmethod
from typing import Any, Iterator, Optional

from ..abstractions.document import BasicDocument
from ..adapters import Adapter
from ..providers.logging import LoggingDatabaseConnection
from ..utils import generate_run_id
from .pipeline import Pipeline


class IngestionPipeline(Pipeline):
    def __init__(
        self,
        adapters: Optional[dict[Any, Adapter]] = None,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs,
    ):
        super().__init__(logging_connection=logging_connection, **kwargs)

    def initialize_pipeline(self, *args, **kwargs) -> None:
        self.pipeline_run_info = {
            "run_id": generate_run_id(),
            "type": "ingestion",
        }

    @property
    @abstractmethod
    def supported_types(self) -> list[str]:
        """
        Returns a list of supported data types.
        """
        pass

    @abstractmethod
    def process_data(
        self, entry_type: str, entry_data: Any
    ) -> Iterator[BasicDocument]:
        """
        Process data into plaintext based on the data type and yield BasicDocument objects.
        """
        pass

    @abstractmethod
    def parse_entry(
        self, entry_type: str, entry_data: Any
    ) -> Iterator[BasicDocument]:
        """
        Parse entry data into plaintext based on the entry type and yield BasicDocument objects.
        """
        pass

    def run(
        self,
        document_id: str,
        blobs: dict[str, Any],
        metadata: Optional[dict] = None,
        **kwargs,
    ) -> Iterator[BasicDocument]:
        """
        Run the appropriate parsing method based on the data type and whether the data is a file or an entry.
        Yields the processed BasicDocument objects.
        """
        self.initialize_pipeline()

        if len(blobs) == 0:
            raise ValueError("No blobs provided to process.")

        for entry_type, blob in blobs.items():
            if entry_type not in self.supported_types:
                raise ValueError(f"IngestionType {entry_type} not supported.")
            yield from self.parse_entry(entry_type, blob)



"""
The `vecs.experimental.adapter.text` module provides adapter steps specifically designed for
handling text data. It provides two main classes, `TextEmbedding` and `ParagraphChunker`.

All public classes, enums, and functions are re-exported by `vecs.adapters` module.
"""

from typing import Any, Dict, Generator, Iterable, Literal, Optional, Tuple

from flupy import flu
from vecs.exc import MissingDependency

from .base import AdapterContext, AdapterStep

TextEmbeddingModel = Literal[
    "all-mpnet-base-v2",
    "multi-qa-mpnet-base-dot-v1",
    "all-distilroberta-v1",
    "all-MiniLM-L12-v2",
    "multi-qa-distilbert-cos-v1",
    "all-MiniLM-L6-v2",
    "multi-qa-MiniLM-L6-cos-v1",
    "paraphrase-multilingual-mpnet-base-v2",
    "paraphrase-albert-small-v2",
    "paraphrase-multilingual-MiniLM-L12-v2",
    "paraphrase-MiniLM-L3-v2",
    "distiluse-base-multilingual-cased-v1",
    "distiluse-base-multilingual-cased-v2",
]


class TextEmbedding(AdapterStep):
    """
    TextEmbedding is an AdapterStep that converts text media into
    embeddings using a specified sentence transformers model.
    """

    def __init__(
        self,
        *,
        model: TextEmbeddingModel,
        batch_size: int = 8,
        use_auth_token: str = None,
    ):
        """
        Initializes the TextEmbedding adapter with a sentence transformers model.

        Args:
            model (TextEmbeddingModel): The sentence transformers model to use for embeddings.
            batch_size (int): The number of records to encode simultaneously.
            use_auth_token (str): The HuggingFace Hub auth token to use for private models.

        Raises:
            MissingDependency: If the sentence_transformers library is not installed.
        """
        try:
            from sentence_transformers import SentenceTransformer as ST
        except ImportError:
            raise MissingDependency(
                "Missing feature vecs[text_embedding]. Hint: `pip install 'vecs[text_embedding]'`"
            )

        self.model = ST(model, use_auth_token=use_auth_token)
        self._exported_dimension = (
            self.model.get_sentence_embedding_dimension()
        )
        self.batch_size = batch_size

    @property
    def exported_dimension(self) -> Optional[int]:
        """
        Returns the dimension of the embeddings produced by the sentence transformers model.

        Returns:
            int: The dimension of the embeddings.
        """
        return self._exported_dimension

    def __call__(
        self,
        records: Iterable[Tuple[str, Any, Optional[Dict]]],
        adapter_context: AdapterContext,  # pyright: ignore
    ) -> Generator[Tuple[str, Any, Dict], None, None]:
        """
        Converts each media in the records to an embedding and yields the result.

        Args:
            records: Iterable of tuples each containing an id, a media and an optional dict.
            adapter_context: Context of the adapter.

        Yields:
            Tuple[str, Any, Dict]: The id, the embedding, and the metadata.
        """
        for batch in flu(records).chunk(self.batch_size):
            batch_records = [x for x in batch]
            media = [text for _, text, _ in batch_records]

            embeddings = self.model.encode(media, normalize_embeddings=True)

            for (id, _, metadata), embedding in zip(batch_records, embeddings):  # type: ignore
                yield (id, embedding, metadata or {})


class ParagraphChunker(AdapterStep):
    """
    ParagraphChunker is an AdapterStep that splits text media into
    paragraphs and yields each paragraph as a separate record.
    """

    def __init__(self, *, skip_during_query: bool):
        """
        Initializes the ParagraphChunker adapter.

        Args:
            skip_during_query (bool): Whether to skip chunking during querying.
        """
        self.skip_during_query = skip_during_query

    def __call__(
        self,
        records: Iterable[Tuple[str, Any, Optional[Dict]]],
        adapter_context: AdapterContext,
    ) -> Generator[Tuple[str, Any, Dict], None, None]:
        """
        Splits each media in the records into paragraphs and yields each paragraph
        as a separate record. If the `skip_during_query` attribute is set to True,
        this step is skipped during querying.

        Args:
            records (Iterable[Tuple[str, Any, Optional[Dict]]]): Iterable of tuples each containing an id, a media and an optional dict.
            adapter_context (AdapterContext): Context of the adapter.

        Yields:
            Tuple[str, Any, Dict]: The id appended with paragraph index, the paragraph, and the metadata.
        """
        if (
            adapter_context == AdapterContext("query")
            and self.skip_during_query
        ):
            for id, media, metadata in records:
                yield (id, media, metadata or {})
        else:
            for id, media, metadata in records:
                paragraphs = media.split("\n\n")

                for paragraph_ix, paragraph in enumerate(paragraphs):
                    yield (
                        f"{id}_para_{str(paragraph_ix).zfill(3)}",
                        paragraph,
                        metadata or {},
                    )



"""A simple example to demonstrate the usage of `WebSearchRAGPipeline`."""

import uvicorn

from r2r.main import E2EPipelineFactory, R2RConfig
from r2r.pipelines import WebSearchRAGPipeline

# Creates a pipeline using the `WebSearchRAGPipeline` implementation
app = E2EPipelineFactory.create_pipeline(
    config=R2RConfig.load_config(), rag_pipeline_impl=WebSearchRAGPipeline
)


if __name__ == "__main__":
    # Run the FastAPI application using Uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



import logging
from dataclasses import dataclass
from typing import Any, Generator, Union

from openai.types.chat import ChatCompletion, ChatCompletionChunk

from r2r.core import GenerationConfig, LLMConfig, LLMProvider

logger = logging.getLogger(__name__)


@dataclass
class LiteLLMConfig(LLMConfig):
    """Configuration for LiteLLM models."""

    provider_name: str = "litellm"


class LiteLLM(LLMProvider):
    """A concrete class for creating LiteLLM models."""

    def __init__(
        self,
        config: LiteLLMConfig,
        *args,
        **kwargs,
    ) -> None:
        logger.info(f"Initializing `LiteLLM` with config: {config}")
        super().__init__()

        if not isinstance(config, LiteLLMConfig):
            raise ValueError(
                "The provided config must be an instance of LiteLLMConfig."
            )
        self.config: LiteLLMConfig = config

        try:
            from litellm import completion

            self.litellm_completion = completion
        except ImportError:
            raise ImportError(
                "Error, `litellm` is required to run a LiteLLM. Please install it using `pip install litellm`."
            )

    def get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletion:
        if generation_config.stream:
            raise ValueError(
                "Stream must be set to False to use the `get_completion` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def get_completion_stream(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> Generator[ChatCompletionChunk, None, None]:
        if not generation_config.stream:
            raise ValueError(
                "Stream must be set to True to use the `get_completion_stream` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def _get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        # Create a dictionary with the default arguments
        args = self._get_base_args(generation_config)
        args["messages"] = messages

        # Conditionally add the 'functions' argument if it's not None
        if generation_config.functions is not None:
            args["functions"] = generation_config.functions

        args = {**args, **kwargs}
        response = self.litellm_completion(**args)

        if not generation_config.stream:
            return ChatCompletion(**response.dict())
        else:
            return self._get_chat_completion(response)

    def _get_chat_completion(
        self,
        response: Any,
    ) -> Generator[ChatCompletionChunk, None, None]:
        for part in response:
            yield ChatCompletionChunk(**part.dict())

    def _get_base_args(
        self,
        generation_config: GenerationConfig,
        prompt=None,
    ) -> dict:
        """Get the base arguments for the LiteLLM API."""
        args = {
            "model": generation_config.model,
            "temperature": generation_config.temperature,
            "top_p": generation_config.top_p,
            "stream": generation_config.stream,
            # TODO - We need to cap this to avoid potential errors when exceed max allowable context
            "max_tokens": generation_config.max_tokens_to_sample,
        }
        return args



from .litellm.base import LiteLLM, LiteLLMConfig
from .llama_cpp.base import LlamaCPP, LlamaCppConfig
from .openai.base import OpenAIConfig, OpenAILLM

__all__ = [
    "LiteLLMConfig",
    "LiteLLM",
    "OpenAIConfig",
    "OpenAILLM",
    "LlamaCPP",
    "LlamaCppConfig",
]



from typing import Optional

from openai.types.chat import ChatCompletion


class RAGPipelineOutput:
    def __init__(
        self,
        search_results: list,
        context: Optional[str] = None,
        completion: Optional[ChatCompletion] = None,
    ):
        self.search_results = search_results
        self.context = context
        self.completion = completion

    def to_dict(self):
        return {
            "search_results": self.search_results,
            "context": self.context,
            "completion": (
                self.completion.to_dict() if self.completion else None
            ),
        }

    def __repr__(self):
        return f"RAGPipelineOutput(search_results={self.search_results}, context={self.context}, completion={self.completion})"



from .base import generate_id_from_label, generate_run_id
from .splitter.text import RecursiveCharacterTextSplitter, TextSplitter

__all__ = [
    "RecursiveCharacterTextSplitter",
    "TextSplitter",
    "generate_run_id",
    "generate_id_from_label",
]



"""
Abstract base class for completion pipelines.
"""

import json
import logging
import uuid
from abc import abstractmethod
from typing import Any, Generator, Optional, Union

from openai.types.chat import ChatCompletion

from ..abstractions.output import RAGPipelineOutput
from ..providers.llm import GenerationConfig, LLMProvider
from ..providers.logging import LoggingDatabaseConnection, log_execution_to_db
from ..providers.prompt import PromptProvider
from .pipeline import Pipeline

logger = logging.getLogger(__name__)


class RAGPipeline(Pipeline):
    SEARCH_STREAM_MARKER = "search"
    CONTEXT_STREAM_MARKER = "context"
    COMPLETION_STREAM_MARKER = "completion"

    def __init__(
        self,
        llm: "LLMProvider",
        prompt_provider: PromptProvider,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs,
    ):
        self.llm = llm
        self.prompt_provider = prompt_provider
        super().__init__(logging_connection=logging_connection, **kwargs)

    def initialize_pipeline(
        self, query: str, search_only: bool, *args, **kwargs
    ) -> None:
        self.pipeline_run_info = {
            "run_id": uuid.uuid4(),
            "type": "rag" if not search_only else "search",
        }
        self.ingress(query)

    @log_execution_to_db
    def ingress(self, data: Any) -> Any:
        """
        Ingresses data into the pipeline.
        """
        self._check_pipeline_initialized()
        return data

    @abstractmethod
    def transform_query(self, query: str) -> Any:
        """
        Transforms the input query for retrieval.
        """
        pass

    @abstractmethod
    def search(
        self,
        transformed_query,
        filters: dict[str, Any],
        limit: int,
        *args,
        **kwargs,
    ) -> list:
        """
        Retrieves results based on the transformed query.
        The search_type parameter allows for specifying the type of search,
        """
        pass

    @abstractmethod
    def rerank_results(self, results: list) -> list:
        """
        Reranks the retrieved results based on relevance or other criteria.
        """
        pass

    @abstractmethod
    def _format_results(self, results: list) -> str:
        """
        Formats the results for generation.
        """
        pass

    @log_execution_to_db
    def construct_context(
        self,
        results: list,
    ) -> str:
        reranked_results = self.rerank_results(results)
        return self._format_results(reranked_results)

    @log_execution_to_db
    def construct_prompt(self, inputs: dict[str, str]) -> str:
        """
        Constructs a prompt for generation based on the reranked chunks.
        """
        return self.prompt_provider.get_prompt("task_prompt", inputs).format(
            **inputs
        )

    @log_execution_to_db
    def generate_completion(
        self,
        prompt: str,
        generation_config: GenerationConfig,
    ) -> Union[Generator[str, None, None], ChatCompletion]:
        """
        Generates a completion based on the prompt.
        """
        self._check_pipeline_initialized()
        messages = [
            {
                "role": "system",
                "content": self.prompt_provider.get_prompt("system_prompt"),
            },
            {
                "role": "user",
                "content": prompt,
            },
        ]
        if not generation_config.stream:
            return self.llm.get_completion(messages, generation_config)

        return self._stream_generate_completion(messages, generation_config)

    def _stream_generate_completion(
        self, messages: list[dict], generation_config: GenerationConfig
    ) -> Generator[str, None, None]:
        for result in self.llm.get_completion_stream(
            messages, generation_config
        ):
            yield result.choices[0].delta.content or ""  # type: ignore

    def run(
        self,
        query,
        filters={},
        limit=10,
        search_only=False,
        generation_config: Optional[GenerationConfig] = None,
        *args,
        **kwargs,
    ) -> Union[Generator[str, None, None], RAGPipelineOutput]:
        """
        Runs the completion pipeline.
        """
        self.initialize_pipeline(query, search_only)

        transformed_query = self.transform_query(query)
        search_results = self.search(transformed_query, filters, limit)
        if search_only:
            return RAGPipelineOutput(search_results, None, None)
        elif not generation_config:
            raise ValueError(
                "GenerationConfig is required for completion generation."
            )
        elif search_only and generation_config:
            raise ValueError(
                "GenerationConfig is not required for search only."
            )

        context = self.construct_context(search_results)

        prompt = self.construct_prompt(
            {"query": transformed_query, "context": context}
        )

        if not generation_config.stream:
            completion = self.generate_completion(prompt, generation_config)
            return RAGPipelineOutput(search_results, context, completion)

        return self._stream_run(
            search_results, context, prompt, generation_config
        )

    def _stream_run(
        self,
        search_results: list,
        context: str,
        prompt: str,
        generation_config: GenerationConfig,
    ) -> Generator[str, None, None]:
        yield f"<{RAGPipeline.SEARCH_STREAM_MARKER}>"
        yield json.dumps([ele.to_dict() for ele in search_results])
        yield f"</{RAGPipeline.SEARCH_STREAM_MARKER}>"

        yield f"<{RAGPipeline.CONTEXT_STREAM_MARKER}>"
        yield context
        yield f"</{RAGPipeline.CONTEXT_STREAM_MARKER}>"
        yield f"<{RAGPipeline.COMPLETION_STREAM_MARKER}>"
        for chunk in self.generate_completion(prompt, generation_config):
            yield chunk
        yield f"</{RAGPipeline.COMPLETION_STREAM_MARKER}>"

    def _get_extra_args(self, *args, **kwargs):
        """
        Retrieves any extra arguments needed for the pipeline's operations.
        """
        return {}



import os
import tempfile
from typing import Optional

import requests

from ..base import Adapter


class ReductoAPIError(Exception):
    """Custom exception class for Reducto API errors."""

    pass


class ReductoAdapter(Adapter[dict]):
    def __init__(
        self, s3_bucket: Optional[str] = None, api_key: Optional[str] = None
    ):
        try:
            import boto3
        except ImportError:
            raise ImportError("Please install boto3 to use the ReductoAdapter")

        api_key = api_key or os.getenv("REDUCTO_API_KEY")
        s3_bucket = s3_bucket or os.getenv("AWS_S3_BUCKET")

        if not api_key:
            raise ValueError(
                "Reducto API key not found. Please set the REDUCTO_API_KEY environment variable."
            )
        if not s3_bucket:
            raise ValueError(
                "AWS S3 bucket name not found. Please set the AWS_S3_BUCKET_NAME environment variable."
            )

        self.api_key = api_key
        self.s3_bucket = s3_bucket
        self.s3 = boto3.client("s3")

    def adapt(self, data: bytes) -> list[str]:
        # Upload the file to S3

        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(data)
            file_path = temp_file.name
            file_name = os.path.basename(file_path)

        self.s3.upload_file(file_path, self.s3_bucket, file_name)
        os.unlink(file_path)

        # Generate the presigned URL
        presigned_url = self.s3.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.s3_bucket, "Key": file_name},
            ExpiresIn=3600,  # URL expires in 1 hour (3600 seconds)
        )

        # Prepare the request payload
        payload = {
            "document_url": presigned_url,
            "config": {
                "chunk_size": None,
                "disable_chunking": False,
                "mode": "document",
                "table_summary": True,
                "figure_summary": False,
                "chart_extract": False,
                "enrich": False,
                "merge_tables": False,
                "dpi": 200,
            },
        }

        # Make the API request
        url = "https://v1.api.reducto.ai/parse"
        headers = {
            "accept": "application/json",
            "authorization": f"Bearer {self.api_key}",
            "content-type": "application/json",
        }

        response = requests.post(url, json=payload, headers=headers)

        # Check the response status code
        if response.status_code == 200:
            response_json = response.json()
            result = response_json["result"]
            for chunk in result["chunks"]:
                yield chunk["content"]
        else:
            raise ReductoAPIError(
                f"Error processing document. Status code: {response.status_code}. Error message: {response.text}"
            )



import json
from abc import ABC, abstractmethod
from typing import Any, Union
from uuid import UUID


class VectorEntry:
    def __init__(
        self, entry_id: UUID, vector: list[float], metadata: dict[str, Any]
    ):
        self.vector = vector
        self.id = entry_id
        self.metadata = metadata

    def to_json(self) -> str:
        """Serialize the object to a JSON string."""
        return json.dumps(
            {"id": self.id, "vector": self.vector, "metadata": self.metadata},
            default=lambda o: o.__dict__,
        )

    @staticmethod
    def from_json(json_str: str) -> "VectorEntry":
        """Deserialize a JSON string into a VectorEntry object."""
        data = json.loads(json_str)
        return VectorEntry(
            vector=data["vector"],
            entry_id=data["id"],
            metadata=data["metadata"],
        )

    def __str__(self) -> str:
        """Return a string representation of the VectorEntry."""
        return f"VectorEntry(id={self.id}, vector={self.vector}, metadata={self.metadata})"

    def __repr__(self) -> str:
        """Return an unambiguous string representation of the VectorEntry."""
        return f"VectorEntry(id={self.id}, vector={self.vector}, metadata={self.metadata})"


class VectorSearchResult:
    def __init__(
        self, entry_id: str, score: float, metadata: dict[str, Any]
    ) -> None:
        self.id = entry_id
        self.score = score
        self.metadata = metadata

    def __str__(self) -> str:
        """Return a string representation of the VectorSearchResult."""
        return f"VectorSearchResult(id={self.id}, score={self.score}, metadata={self.metadata})"

    def __repr__(self) -> str:
        """Return an unambiguous string representation of the VectorSearchResult for debugging."""
        return f"VectorSearchResult(id={self.id}, score={self.score}, metadata={self.metadata})"

    def to_dict(self) -> dict:
        """Return a dictionary representation of the VectorSearchResult."""
        return {
            "id": self.id,
            "score": self.score,
            "metadata": self.metadata,
        }


class VectorDBProvider(ABC):
    supported_providers = ["local", "pgvector", "qdrant"]

    def __init__(self, provider: str):
        if provider not in VectorDBProvider.supported_providers:
            raise ValueError(
                f"Error, `{provider}` is not in VectorDBProvider's list of supported providers."
            )

    @abstractmethod
    def initialize_collection(
        self, collection_name: str, dimension: int
    ) -> None:
        pass

    @abstractmethod
    def copy(self, entry: VectorEntry, commit: bool = True) -> None:
        pass

    @abstractmethod
    def upsert(self, entry: VectorEntry, commit: bool = True) -> None:
        pass

    @abstractmethod
    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        pass

    @abstractmethod
    def create_index(self, index_type, column_name, index_options):
        pass

    @abstractmethod
    def close(self):
        pass

    def upsert_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        for entry in entries:
            self.upsert(entry, commit=commit)

    def copy_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        for entry in entries:
            self.copy(entry, commit=commit)

    @abstractmethod
    def filtered_deletion(
        self, key: str, value: Union[bool, int, str]
    ) -> None:
        pass

    @abstractmethod
    def get_all_unique_values(
        self, metadata_field: str, filters: dict = {}
    ) -> list:
        pass



import logging
import os
from typing import Optional, Union

from r2r.core import VectorDBProvider, VectorEntry, VectorSearchResult

logger = logging.getLogger(__name__)


class QdrantDB(VectorDBProvider):
    def __init__(self, provider: str = "qdrant") -> None:
        logger.info(
            "Initializing `QdrantDB` to store and retrieve embeddings."
        )

        super().__init__(provider)
        if provider != "qdrant":
            raise ValueError(
                "QdrantDB must be initialized with provider `qdrant`."
            )
        try:
            from qdrant_client import QdrantClient
            from qdrant_client.http import models

            self.models = models
        except ImportError:
            raise ValueError(
                f"Error, `qdrant_client` is not installed. Please install it using `pip install qdrant-client`."
            )
        try:
            host = os.getenv("QDRANT_HOST")
            port = os.getenv("QDRANT_PORT")
            api_key = os.getenv("QDRANT_API_KEY")

            if not host or not port or not api_key:
                raise ValueError(
                    "Error, QdrantDB requires the QDRANT_HOST, QDRANT_PORT, and QDRANT_API_KEY environment variables."
                )

            self.client = QdrantClient(host, port=int(port), api_key=api_key)
        except Exception as e:
            raise ValueError(
                f"Error {e} occurred while attempting to connect to the qdrant provider."
            )
        self.collection_name: Optional[str] = None

    def initialize_collection(
        self, collection_name: str, dimension: int
    ) -> None:
        self.collection_name = collection_name
        try:
            result = self.client.create_collection(
                collection_name=f"{collection_name}",
                vectors_config=self.models.VectorParams(
                    size=dimension, distance=self.models.Distance.COSINE
                ),
            )
            if result is False:
                raise ValueError(
                    f"Error occurred while attempting to create collection {collection_name}."
                )
        except Exception:
            # TODO - Handle more appropriately - create collection fails when it already exists
            pass

    def copy(self, entry: VectorEntry, commit=True) -> None:
        raise NotImplementedError(
            "QdrantDB does not support the `copy` method."
        )

    def upsert(self, entry: VectorEntry, commit=True) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert`."
            )
        points = [
            self.models.PointStruct(
                id=str(entry.id),
                vector=list([float(ele) for ele in entry.vector]),
                payload=entry.metadata,
            )
        ]
        self.client.upsert(
            collection_name=f"{self.collection_name}",
            points=points,
        )

    def upsert_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `upsert_entries`."
            )
        points = [
            self.models.PointStruct(
                id=str(entry.id),
                vector=list([float(ele) for ele in entry.vector]),
                payload=entry.metadata,
            )
            for entry in entries
        ]
        self.client.upsert(
            collection_name=f"{self.collection_name}",
            points=points,
        )

    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        if self.collection_name is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `search`."
            )

        results = self.client.search(
            collection_name=f"{self.collection_name}",
            query_filter=self.models.Filter(
                must=[
                    self.models.FieldCondition(
                        key=key,
                        match=self.models.MatchValue(
                            value=value,
                        ),
                    )
                    for key, value in filters.items()
                ]
            ),
            query_vector=query_vector,
            limit=limit,
        )

        return [
            VectorSearchResult(
                str(result.id), result.score, result.payload or {}
            )
            for result in results
        ]

    def create_index(self, index_type, column_name, index_options):
        pass

    def close(self):
        pass

    def filtered_deletion(
        self, key: str, value: Union[bool, int, str]
    ) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `filtered_deletion`."
            )

        self.client.delete(
            collection_name=self.collection_name,
            points_selector=self.models.FilterSelector(
                filter=self.models.Filter(
                    must=[
                        self.models.FieldCondition(
                            key=key,
                            match=self.models.MatchValue(value=value),
                        ),
                    ],
                )
            ),
        )
        return

    def get_all_unique_values(
        self, collection_name: str, metadata_field: str, filters: dict = {}
    ) -> list:
        if self.collection_name is None:
            raise ValueError(
                "Please call `initialize_collection` before attempting to run `get_all_unique_values`."
            )

        # Create a scroll filter based on the provided filters
        scroll_filter = None
        if filters:
            filter_conditions = [
                self.models.FieldCondition(
                    key=key, match=self.models.MatchValue(value=value)
                )
                for key, value in filters.items()
            ]
            scroll_filter = self.models.Filter(must=filter_conditions)

        unique_values = set()

        # Scroll through the collection and retrieve points in batches
        next_page_offset = None
        while True:
            records, next_page_offset = self.scroll(
                collection_name=collection_name,
                scroll_filter=scroll_filter,
                offset=next_page_offset,
                limit=100,  # Adjust the batch size as needed
                with_payload=True,
            )

            for record in records:
                if metadata_field in record.payload:
                    unique_values.add(record.payload[metadata_field])

            if next_page_offset is None:
                break

        return list(unique_values)



import argparse
import os

import uvicorn

from r2r.main import E2EPipelineFactory, R2RConfig

current_file_path = os.path.dirname(__file__)
configs_path = os.path.join(current_file_path, "..", "configs")

OPTIONS = {
    "default": None,
    "local_ollama": os.path.join(configs_path, "local_ollama.json"),
    "local_llama_cpp": os.path.join(configs_path, "local_llama_cpp.json"),
}


def create_app(config_name: str = "default"):
    config_path = OPTIONS[config_name]

    app = E2EPipelineFactory.create_pipeline(
        config=R2RConfig.load_config(config_path)
    )
    return app


app = create_app()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="R2R Pipeline")
    parser.add_argument(
        "--config",
        type=str,
        default="default",
        choices=OPTIONS.keys(),
        help="Configuration option for the pipeline",
    )
    args, _ = parser.parse_known_args()

    uvicorn.run(app, host="0.0.0.0", port=8000)



from abc import ABC, abstractmethod
from typing import Any, Optional


class PromptProvider(ABC):
    @abstractmethod
    def add_prompt(self, prompt_name: str, prompt: str) -> None:
        pass

    @abstractmethod
    def get_prompt(
        self, prompt_name: str, inputs: Optional[dict[str, Any]] = None
    ) -> str:
        pass

    @abstractmethod
    def get_all_prompts(self) -> dict[str, str]:
        pass


class DefaultPromptProvider(PromptProvider):
    def __init__(self) -> None:
        self.prompts: dict[str, str] = {}

    def add_prompt(self, prompt_name: str, prompt: str) -> None:
        self.prompts[prompt_name] = prompt

    def get_prompt(
        self, prompt_name: str, inputs: Optional[dict[str, Any]] = None
    ) -> str:
        prompt = self.prompts.get(prompt_name)
        if prompt is None:
            raise ValueError(f"Prompt '{prompt_name}' not found.")
        return prompt.format(**(inputs or {}))

    def set_prompt(self, prompt_name: str, prompt: str) -> None:
        if prompt_name not in self.prompts:
            raise ValueError(f"Prompt '{prompt_name}' not found.")
        self.prompts[prompt_name] = prompt

    def get_all_prompts(self) -> dict[str, str]:
        return self.prompts.copy()



import logging
from typing import Any

import dotenv

from r2r.core import LoggingDatabaseConnection
from r2r.core.utils import RecursiveCharacterTextSplitter
from r2r.llms import (
    LiteLLM,
    LiteLLMConfig,
    LlamaCPP,
    LlamaCppConfig,
    OpenAIConfig,
    OpenAILLM,
)
from r2r.pipelines import (
    BasicEmbeddingPipeline,
    BasicEvalPipeline,
    BasicIngestionPipeline,
    BasicRAGPipeline,
)

from .app import create_app
from .utils import R2RConfig

dotenv.load_dotenv()


class E2EPipelineFactory:
    @staticmethod
    def get_vector_db(database_config: dict[str, Any]):
        if database_config["provider"] == "qdrant":
            from r2r.vector_dbs import QdrantDB

            return QdrantDB()
        elif database_config["provider"] == "pgvector":
            from r2r.vector_dbs import PGVectorDB

            return PGVectorDB()
        elif database_config["provider"] == "local":
            from r2r.vector_dbs import LocalVectorDB

            return LocalVectorDB()

    @staticmethod
    def get_embeddings_provider(embedding_config: dict[str, Any]):
        if embedding_config["provider"] == "openai":
            from r2r.embeddings import OpenAIEmbeddingProvider

            return OpenAIEmbeddingProvider()
        elif embedding_config["provider"] == "sentence-transformers":
            from r2r.embeddings import SentenceTransformerEmbeddingProvider

            return SentenceTransformerEmbeddingProvider(
                embedding_config["model"]
            )
        else:
            raise ValueError(
                f"Embedding provider {embedding_config['provider']} not supported"
            )

    @staticmethod
    def get_llm(llm_config: dict[str, Any]):
        if llm_config["provider"] == "openai":
            return OpenAILLM(OpenAIConfig())
        elif llm_config["provider"] == "litellm":
            return LiteLLM(LiteLLMConfig())
        elif llm_config["provider"] == "llama-cpp":
            return LlamaCPP(
                LlamaCppConfig(
                    llm_config.get("model_path", ""),
                    llm_config.get("model_name", ""),
                )
            )

    @staticmethod
    def get_text_splitter(text_splitter_config: dict[str, Any]):
        if text_splitter_config["type"] != "recursive_character":
            raise ValueError(
                "Only recursive character text splitter is supported"
            )
        return RecursiveCharacterTextSplitter(
            chunk_size=text_splitter_config["chunk_size"],
            chunk_overlap=text_splitter_config["chunk_overlap"],
            length_function=len,
            is_separator_regex=False,
        )

    @staticmethod
    def create_pipeline(
        config: R2RConfig,
        db=None,
        embeddings_provider=None,
        llm=None,
        text_splitter=None,
        adapters=None,
        ingestion_pipeline_impl=BasicIngestionPipeline,
        embedding_pipeline_impl=BasicEmbeddingPipeline,
        rag_pipeline_impl=BasicRAGPipeline,
        eval_pipeline_impl=BasicEvalPipeline,
        app_fn=create_app,
    ):
        logging.basicConfig(level=config.logging_database["level"])

        embeddings_provider = (
            embeddings_provider
            or E2EPipelineFactory.get_embeddings_provider(config.embedding)
        )
        embedding_model = config.embedding["model"]
        embedding_dimension = config.embedding["dimension"]
        embedding_batch_size = config.embedding["batch_size"]

        db = db or E2EPipelineFactory.get_vector_db(config.vector_database)
        collection_name = config.vector_database["collection_name"]
        db.initialize_collection(collection_name, embedding_dimension)

        llm = llm or E2EPipelineFactory.get_llm(config.language_model)

        logging_connection = LoggingDatabaseConnection(
            config.logging_database["provider"],
            config.logging_database["collection_name"],
        )

        cmpl_pipeline = rag_pipeline_impl(
            llm,
            db=db,
            embedding_model=embedding_model,
            embeddings_provider=embeddings_provider,
            logging_connection=logging_connection,
        )

        text_splitter = text_splitter or E2EPipelineFactory.get_text_splitter(
            config.ingestion["text_splitter"]
        )

        embd_pipeline = embedding_pipeline_impl(
            embedding_model,
            embeddings_provider,
            db,
            logging_connection=logging_connection,
            text_splitter=text_splitter,
            embedding_batch_size=embedding_batch_size,
        )

        eval_pipeline = eval_pipeline_impl(
            config.evals, logging_connection=logging_connection
        )
        ingst_pipeline = ingestion_pipeline_impl(adapters=adapters)

        app = app_fn(
            ingestion_pipeline=ingst_pipeline,
            embedding_pipeline=embd_pipeline,
            rag_pipeline=cmpl_pipeline,
            eval_pipeline=eval_pipeline,
            config=config,
            logging_connection=logging_connection,
        )

        return app



import os

from r2r.core import EvalProvider


class DeepEvalProvider(EvalProvider):
    def __init__(self, provider: str, sampling_fraction: float = 1.0):
        super().__init__(provider, sampling_fraction)
        try:
            from deepeval import evaluate
            from deepeval.metrics import (
                AnswerRelevancyMetric,
                HallucinationMetric,
            )
            from deepeval.test_case import LLMTestCase

            self.deep_evaluate = evaluate
            self.AnswerRelevancyMetric = AnswerRelevancyMetric
            self.HallucinationMetric = HallucinationMetric
            self.LLMTestCase = LLMTestCase
        except ImportError:
            raise ImportError(
                "DeepEval is not installed. Please install it using `pip install deepeval`."
            )
        if not os.getenv("OPENAI_API_KEY"):
            raise ValueError(
                "Please set the `OPENAI_API_KEY` environment variable to run with DeepEval."
            )

    def _evaluate(
        self, query: str, context: str, completion: str
    ) -> dict[str, dict[str, str]]:
        test_case = self.LLMTestCase(
            input=query,
            actual_output=completion,
            context=[context],
            retrieval_context=[context],
        )
        # TODO - Make inner metrics configurable.

        answer_relevancy_metric = self.AnswerRelevancyMetric()
        hallucination_metric = self.HallucinationMetric()

        answer_relevancy_result = self.deep_evaluate(
            [test_case], [answer_relevancy_metric]
        )
        hallucination_result = self.deep_evaluate(
            [test_case], [hallucination_metric]
        )

        # TODO - Fix return type across evals, or at least locally.
        return {
            "answer_relevancy": {
                "score": answer_relevancy_result[0].metrics[0].score,
                "reason": answer_relevancy_result[0].metrics[0].reason,
            },
            "hallucination": {
                "score": hallucination_result[0].metrics[0].score,
                "reason": hallucination_result[0].metrics[0].reason,
            },
        }



"""
A simple example to demonstrate the usage of `BasicIngestionPipeline`.
"""

import logging
from enum import Enum
from typing import Any, Iterator, Optional, Union

from r2r.core import (
    BasicDocument,
    IngestionPipeline,
    LoggingDatabaseConnection,
)
from r2r.core.adapters import (
    Adapter,
    HTMLAdapter,
    JSONAdapter,
    PDFAdapter,
    TextAdapter,
)

logger = logging.getLogger(__name__)


class IngestionType(Enum):
    TXT = "txt"
    JSON = "json"
    HTML = "html"
    PDF = "pdf"


class BasicIngestionPipeline(IngestionPipeline):
    """
    Processes incoming documents into plaintext based on their data type.
    Supports TXT, JSON, HTML, and PDF formats.
    """

    def __init__(
        self,
        adapters: Optional[dict[IngestionType, Adapter]] = None,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
    ):
        logger.info(
            f"Initializing a `BasicIngestionPipeline` to process incoming documents."
        )

        super().__init__(
            logging_connection,
        )
        self.pipeline_run_info = None
        self.default_adapters = {
            IngestionType.TXT: TextAdapter(),
            IngestionType.JSON: JSONAdapter(),
            IngestionType.HTML: HTMLAdapter(),
            IngestionType.PDF: PDFAdapter(),
        }
        self.adapters = self.default_adapters
        if adapters is not None:
            for entry_type, adapter in adapters.items():
                self.adapters[entry_type] = adapter

    @property
    def supported_types(self) -> list[str]:
        """
        Lists the data types supported by the pipeline.
        """
        return [entry_type.value for entry_type in IngestionType]

    def process_data(
        self,
        entry_type: IngestionType,
        entry_data: Union[bytes, str],
    ) -> Iterator[BasicDocument]:
        adapter = self.adapters.get(
            entry_type, self.default_adapters[entry_type]
        )
        texts = adapter.adapt(entry_data)
        for text in texts:
            yield BasicDocument(
                id=self.document_id, text=text, metadata=self.metadata
            )

    def parse_entry(
        self, entry_type: str, entry_data: Union[bytes, str]
    ) -> Iterator[BasicDocument]:
        yield from self.process_data(IngestionType(entry_type), entry_data)

    def run(
        self,
        document_id: str,
        blobs: dict[str, Any],
        metadata: Optional[dict] = None,
        **kwargs,
    ) -> Iterator[BasicDocument]:
        self.initialize_pipeline()
        self.document_id = document_id
        self.metadata = metadata or {}

        if len(blobs) == 0:
            raise ValueError("No blobs provided to process.")

        for entry_type, blob in blobs.items():
            if entry_type not in self.supported_types:
                raise ValueError(f"IngestionType {entry_type} not supported.")
            yield from self.parse_entry(entry_type, blob)



import json
import string
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, Generic, TypeVar, Union

from bs4 import BeautifulSoup

T = TypeVar("T")


class Adapter(ABC, Generic[T]):
    @abstractmethod
    def adapt(self, data: Any) -> T:
        pass


class TextAdapter(Adapter[str]):
    def adapt(self, data: str) -> list[str]:
        return [data]


class JSONAdapter(Adapter[str]):
    def adapt(self, data: Union[str, bytes]) -> list[str]:
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        return [self._parse_json(json.loads(data))]

    def _parse_json(self, data: dict) -> str:
        def remove_objects_with_null(obj):
            if not isinstance(obj, dict):
                return obj
            result = obj.copy()
            for key, value in obj.items():
                if isinstance(value, dict):
                    result[key] = remove_objects_with_null(value)
                elif value is None:
                    del result[key]
            return result

        def format_json_as_text(obj, indent=0):
            lines = []
            indent_str = " " * indent

            if isinstance(obj, dict):
                for key, value in obj.items():
                    if isinstance(value, (dict, list)):
                        nested = format_json_as_text(value, indent + 2)
                        lines.append(f"{indent_str}{key}:\n{nested}")
                    else:
                        lines.append(f"{indent_str}{key}: {value}")
            elif isinstance(obj, list):
                for item in obj:
                    nested = format_json_as_text(item, indent + 2)
                    lines.append(f"{nested}")
            else:
                return f"{indent_str}{obj}"

            return "\n".join(lines)

        return format_json_as_text(remove_objects_with_null(data))


class HTMLAdapter(Adapter[str]):
    def adapt(self, data: str) -> list[str]:
        soup = BeautifulSoup(data, "html.parser")
        return [soup.get_text()]


class PDFAdapter(Adapter[str]):
    def __init__(self):
        try:
            from pypdf import PdfReader

            self.PdfReader = PdfReader
        except ImportError:
            raise ValueError(
                "Error, `pypdf` is required to run `PyPDFAdapter`. Please install it using `pip install pypdf`."
            )

    def adapt(self, data: bytes) -> list[str]:
        pdf = self.PdfReader(BytesIO(data))
        text = ""
        for page in pdf.pages:
            page_text = page.extract_text()
            if page_text is not None:
                page_text = "".join(
                    filter(lambda x: x in string.printable, page_text)
                )
                text += page_text + " "
        return [text]



"""
The `vecs.experimental.adapter.base` module provides abstract classes and utilities
for creating and handling adapters in vecs. Adapters allow users to interact with
a collection using media types other than vectors.

All public classes, enums, and functions are re-exported by `vecs.adapters` module.
"""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple

from vecs.exc import ArgError


class AdapterContext(str, Enum):
    """
    An enum representing the different contexts in which a Pipeline
    will be invoked.

    Attributes:
        upsert (str): The Collection.upsert method
        query (str): The Collection.query method
    """

    upsert = "upsert"
    query = "query"


class AdapterStep(ABC):
    """
    Abstract class representing a step in the adapter pipeline.

    Each adapter step should adapt a user media into a tuple of:
     - id (str)
     - media (unknown type)
     - metadata (dict)

    If the user provides id or metadata, default production is overridden.
    """

    @property
    def exported_dimension(self) -> Optional[int]:
        """
        Property that should be overridden by subclasses to provide the output dimension
        of the adapter step.
        """
        return None

    @abstractmethod
    def __call__(
        self,
        records: Iterable[Tuple[str, Any, Optional[Dict]]],
        adapter_context: AdapterContext,
    ) -> Generator[Tuple[str, Any, Dict], None, None]:
        """
        Abstract method that should be overridden by subclasses to handle each record.
        """


class Adapter:
    """
    Class representing a sequence of AdapterStep instances forming a pipeline.
    """

    def __init__(self, steps: List[AdapterStep]):
        """
        Initialize an Adapter instance with a list of AdapterStep instances.

        Args:
            steps: List of AdapterStep instances.

        Raises:
            ArgError: Raised if the steps list is empty.
        """
        self.steps = steps
        if len(steps) < 1:
            raise ArgError("Adapter must contain at least 1 step")

    @property
    def exported_dimension(self) -> Optional[int]:
        """
        The output dimension of the adapter. Returns the exported dimension of the last
        AdapterStep that provides one (from end to start of the steps list).
        """
        for step in reversed(self.steps):
            step_dim = step.exported_dimension
            if step_dim is not None:
                return step_dim
        return None

    def __call__(
        self,
        records: Iterable[Tuple[str, Any, Optional[Dict]]],
        adapter_context: AdapterContext,
    ) -> Generator[Tuple[str, Any, Dict], None, None]:
        """
        Invokes the adapter pipeline on an iterable of records.

        Args:
            records: Iterable of tuples each containing an id, a media and an optional dict.
            adapter_context: Context of the adapter.

        Yields:
            Tuples each containing an id, a media and a dict.
        """
        pipeline = records
        for step in self.steps:
            pipeline = step(pipeline, adapter_context)

        yield from pipeline  # type: ignore



import json
import logging
import os
import sqlite3
from typing import Optional, Union

from r2r.core import VectorDBProvider, VectorEntry, VectorSearchResult

logger = logging.getLogger(__name__)


class LocalVectorDB(VectorDBProvider):
    def __init__(
        self, provider: str = "local", db_path: Optional[str] = None
    ) -> None:
        logger.info(
            "Initializing `LocalVectorDB` to store and retrieve embeddings."
        )

        super().__init__(provider)
        if provider != "local":
            raise ValueError(
                "LocalVectorDB must be initialized with provider `local`."
            )

        self.db_path = db_path
        self.collection_name: Optional[str] = None

    def _get_conn(self):
        conn = sqlite3.connect(self.db_path or os.getenv("LOCAL_DB_PATH"))
        return conn

    def _get_cursor(self, conn):
        return conn.cursor()

    def initialize_collection(
        self, collection_name: str, dimension: int
    ) -> None:
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(
            f"""
            CREATE TABLE IF NOT EXISTS "{collection_name}" (
                id TEXT PRIMARY KEY,
                vector TEXT,
                metadata TEXT
            )
        """
        )
        self.collection_name = collection_name
        conn.commit()
        conn.close()

    def create_index(self, index_type, column_name, index_options):
        raise NotImplementedError(
            "LocalVectorDB does not support creating indexes."
        )

    def copy(self, entry: VectorEntry, commit=True) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(
            f"""
                INSERT OR IGNORE INTO "{self.collection_name}" (id, vector, metadata)
                VALUES (?, ?, ?)
            """,
            (
                str(entry.id),
                json.dumps(entry.vector),
                json.dumps(entry.metadata),
            ),
        )
        if commit:
            conn.commit()
        conn.close()

    def upsert(self, entry: VectorEntry, commit=True) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(
            f"""
            INSERT OR REPLACE INTO "{self.collection_name}" (id, vector, metadata)
            VALUES (?, ?, ?)
        """,
            (
                str(entry.id),
                json.dumps(entry.vector),
                json.dumps(entry.metadata),
            ),
        )
        if commit:
            conn.commit()
        conn.close()

    def _cosine_similarity(
        self, vec1: list[float], vec2: list[float]
    ) -> float:
        dot_product = sum(a * b for a, b in zip(vec1, vec2))
        norm_a = sum(a * a for a in vec1) ** 0.5
        norm_b = sum(b * b for b in vec2) ** 0.5
        return dot_product / (norm_a * norm_b)

    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        if self.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT * FROM "{self.collection_name}"')
        results = []
        for id, vector, metadata in cursor.fetchall():
            vector = json.loads(vector)
            metadata = json.loads(metadata)
            if all(metadata.get(k) == v for k, v in filters.items()):
                # Local cosine similarity calculation
                score = self._cosine_similarity(query_vector, vector)
                results.append(VectorSearchResult(id, score, metadata))
        results.sort(key=lambda x: x.score, reverse=True)
        conn.close()
        return results[:limit]

    def filtered_deletion(
        self, key: str, value: Union[bool, int, str]
    ) -> None:
        if self.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )

        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT * FROM "{self.collection_name}"')
        for id, vector, metadata in cursor.fetchall():
            metadata = json.loads(metadata)
            if metadata.get(key) == value:
                cursor.execute(
                    f'DELETE FROM "{self.collection_name}" WHERE id = ?', (id,)
                )
        conn.commit()
        conn.close()

    def get_all_unique_values(
        self, metadata_field: str, filters: dict = {}
    ) -> list:
        if self.collection_name is None:
            raise ValueError(
                "Collection name is not set. Please call `initialize_collection` first."
            )
        conn = self._get_conn()
        cursor = self._get_cursor(conn)
        cursor.execute(f'SELECT metadata FROM "{self.collection_name}"')
        unique_values = set()
        for (metadata,) in cursor.fetchall():
            metadata = json.loads(metadata)
            if all(metadata.get(k) == v for k, v in filters.items()):
                if metadata_field in metadata:
                    unique_values.add(metadata[metadata_field])
        conn.close()
        return list(unique_values)

    def close(self):
        pass



import logging
import random
from abc import abstractmethod
from typing import Any, Optional

from ..providers.logging import LoggingDatabaseConnection
from ..utils import generate_run_id
from .pipeline import Pipeline

logger = logging.getLogger(__name__)


class EvalPipeline(Pipeline):
    def __init__(
        self,
        frequency: int,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs,
    ):
        self.frequency = frequency
        super().__init__(logging_connection=logging_connection, **kwargs)

    def initialize_pipeline(
        self, run_id: Optional[str], *args, **kwargs
    ) -> None:
        self.pipeline_run_info = {
            "run_id": run_id or generate_run_id(),
            "type": "evaluation",
        }

    @abstractmethod
    def evaluate(self, query: str, context: str, completion: str) -> Any:
        pass

    def run(
        self,
        query: str,
        context: str,
        completion: str,
        run_id: Optional[str],
        **kwargs,
    ):
        self.initialize_pipeline(run_id)
        logger.debug(
            f"Running the `EvaluationPipeline` with id={self.pipeline_run_info}."
        )
        if random.random() < self.frequency:
            return self.evaluate(query, context, completion)
        return None



import random
from abc import ABC, abstractmethod
from typing import Optional, Union


class EvalProvider(ABC):
    providers = ["deepeval", "parea", "none"]

    def __init__(self, provider: str, sampling_fraction: float = 1.0):
        if provider not in self.providers:
            raise ValueError(f"Provider {provider} not supported.")
        if provider == "none" and sampling_fraction != 0.0:
            raise ValueError(
                f"Sampling fraction must be 0.0 when setting evaluation provider to None."
            )

        self.provider = provider
        self.sampling_fraction = sampling_fraction

    def evaluate(
        self, query: str, context: str, completion: str
    ) -> Optional[dict]:
        if random.random() < self.sampling_fraction:
            return self._evaluate(query, context, completion)
        return None

    @abstractmethod
    def _evaluate(
        self, query: str, context: str, completion: str
    ) -> dict[str, dict[str, Union[str, float]]]:
        pass



from .openai.base import OpenAIEmbeddingProvider
from .setence_transformer.base import SentenceTransformerEmbeddingProvider

__all__ = ["OpenAIEmbeddingProvider", "SentenceTransformerEmbeddingProvider"]



"""
A simple example to demonstrate the usage of `BasicRAGPipeline`.
"""

import logging
from typing import Optional

from r2r.core import (
    LLMProvider,
    LoggingDatabaseConnection,
    PromptProvider,
    RAGPipeline,
    VectorDBProvider,
    VectorSearchResult,
    log_execution_to_db,
)
from r2r.embeddings import OpenAIEmbeddingProvider

from .prompt_provider import BasicPromptProvider

logger = logging.getLogger(__name__)


class BasicRAGPipeline(RAGPipeline):
    """
    Implements a basic Retrieve-And-Generate (RAG) pipeline for document retrieval and generation.
    """

    def __init__(
        self,
        llm: LLMProvider,
        db: VectorDBProvider,
        embedding_model: str,
        embeddings_provider: OpenAIEmbeddingProvider,
        prompt_provider: Optional[PromptProvider] = None,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
    ) -> None:
        """
        Initializes the RAG pipeline with necessary components and configurations.
        """
        logger.debug(f"Initalizing `BasicRAGPipeline`.")
        if not prompt_provider:
            prompt_provider = BasicPromptProvider()
        self.prompt_provider = prompt_provider

        super().__init__(
            llm,
            prompt_provider=prompt_provider,
            logging_connection=logging_connection,
        )
        self.embedding_model = embedding_model
        self.embeddings_provider = embeddings_provider
        self.db = db
        self.pipeline_run_info = None

    def transform_query(self, query: str) -> str:
        """
        Transforms the input query before retrieval, if necessary.
        """
        self._check_pipeline_initialized()
        return query

    @log_execution_to_db
    def search(
        self,
        transformed_query: str,
        filters: dict,
        limit: int,
        *args,
        **kwargs,
    ) -> list[VectorSearchResult]:
        """
        Searches the vector database with the transformed query to retrieve relevant documents.
        """
        logger.debug(f"Retrieving results for query: {transformed_query}")
        self._check_pipeline_initialized()
        results = self.db.search(
            query_vector=self.embeddings_provider.get_embedding(
                transformed_query,
                self.embedding_model,
            ),
            filters=filters,
            limit=limit,
        )
        logger.debug(f"Retrieved the raw results shown:\n{results}\n")

        return results

    def rerank_results(
        self, results: list[VectorSearchResult]
    ) -> list[VectorSearchResult]:
        """
        Reranks the retrieved documents based on relevance, if necessary.
        """
        self._check_pipeline_initialized()
        # Placeholder for reranking logic - A unit transformation.
        return results

    def _format_results(self, results: list[VectorSearchResult]) -> str:
        """
        Formats the reranked results into a human-readable string.
        """
        return "\n\n".join([ele.metadata["text"] for ele in results])



"""A module for creating LlamaCpp model abstractions."""

import datetime
import logging
import os
from dataclasses import dataclass
from typing import Union

from openai.types.chat import ChatCompletion, ChatCompletionChunk

from r2r.core import GenerationConfig, LLMConfig, LLMProvider

logger = logging.getLogger(__name__)


@dataclass
class LlamaCppConfig(LLMConfig):
    """Configuration for LlamaCpp models."""

    # Base
    provider_name: str = "llama-cpp"
    model_path: str = ""
    model_name: str = ""

    def __post_init__(self):
        if not self.model_path or self.model_path == "":
            self.model_path = os.path.join(
                os.path.expanduser("~"), ".cache", "models"
            )
        if not self.model_name or self.model_name == "":
            self.model_name = "tinyllama-1.1b-chat-v1.0.Q2_K.gguf"


class LlamaCPP(LLMProvider):
    """A concrete class for creating LlamaCpp models."""

    def __init__(
        self,
        config: LlamaCppConfig,
        *args,
        **kwargs,
    ) -> None:
        logger.info(f"Initializing `LlamaCPP` with config: {config}")
        super().__init__()

        if not isinstance(config, LlamaCppConfig):
            raise ValueError(
                "The provided config must be an instance of LlamaCppConfig."
            )

        self.config: LlamaCppConfig = config

        try:
            from llama_cpp import Llama
        except ImportError:
            raise ImportError(
                "Error, `llama-cpp-python` is required to run a LlamaCPP. Please install it using `pip install llama-cpp-python`."
            )

        path = os.path.join(self.config.model_path, self.config.model_name)
        self.client = Llama(path, n_ctx=2048)

    def get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletion:
        if generation_config.stream:
            raise ValueError(
                "Stream must be set to False to use the `get_completion` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def get_completion_stream(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletionChunk:
        if not generation_config.stream:
            raise ValueError(
                "Stream must be set to True to use the `get_completion_stream` method."
            )
        return self._get_completion(messages, generation_config, **kwargs)

    def _get_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> Union[ChatCompletion, ChatCompletionChunk]:
        """Get a completion from the LlamaCpp model based on the provided messages."""

        args = self._get_base_args(generation_config)

        prompt = "\n".join([msg["content"] for msg in messages])

        response = self.client(prompt, **args)

        if not generation_config.stream:
            return ChatCompletion(
                # TODO - Set an intelligent id
                id="777",
                object="chat.completion",
                created=int(datetime.datetime.now().timestamp()),
                model=generation_config.model,
                choices=[
                    {
                        "message": {
                            "role": "assistant",
                            "content": str(response),
                        },
                        "index": 0,
                        "finish_reason": "stop",
                    }
                ],
            )
        else:
            return ChatCompletionChunk(
                choices=[
                    {
                        "message": {
                            "role": "assistant",
                            "content": str(response),
                        },
                        "index": 0,
                        "finish_reason": None,
                    }
                ],
                **kwargs,
            )

    def _get_base_args(
        self,
        generation_config: GenerationConfig,
    ) -> dict:
        """Get the base arguments for the LlamaCpp model."""
        args = {
            "temperature": generation_config.temperature,
            "top_p": generation_config.top_p,
            "max_tokens": generation_config.max_tokens_to_sample,
        }
        return args



import glob
import os

from r2r.client import R2RClient
from r2r.core.utils import generate_id_from_label

# Initialize the client with the base URL of your API
base_url = "http://localhost:8000"  # Change this to your actual API base URL
client = R2RClient(base_url)

titles = {
    "meditations.pdf": "Title: Meditations - Marcus Aurelius",
    "the_republic.pdf": "Title: The Republic - Plato",
    "test.pdf": "Title: A Test Document",
}

user_id_0 = generate_id_from_label("user_0")

current_file_directory = os.path.dirname(os.path.abspath(__file__))
data_path = os.path.join(current_file_directory, "..", "data")
print("data_path = ", data_path)
# Use this directory in the glob pattern
for file_path in glob.glob(os.path.join(data_path, "*.pdf")):
    file_name = file_path.split(os.path.sep)[-1]
    print(f"Uploading and processing file: {file_path}")
    # # Upload and process a file
    document_id = str(generate_id_from_label(file_path))
    metadata = {"user_id": user_id_0, "chunk_prefix": titles[file_name]}
    settings: dict = {}
    upload_response = client.upload_and_process_file(
        document_id, file_path, metadata, settings
    )

prompt = """You are given a user query {query} and a user context {context}. Use the context to answer the given query. """
formatted_prompt = prompt.format(
    query="What are the key themes of these books?",
    context="User Uploads:\nTitle: Meditations - Marcus Aurelius\nTitle: The Republic - Plato",
)
# Perform a search on this file
search_response = client.rag_completion(
    formatted_prompt, 5, filters={"user_id": user_id_0}
)
print(search_response)



from .abstractions.document import BasicDocument
from .abstractions.output import RAGPipelineOutput
from .pipelines.embedding import EmbeddingPipeline
from .pipelines.eval import EvalPipeline
from .pipelines.ingestion import IngestionPipeline
from .pipelines.rag import RAGPipeline
from .providers.embedding import EmbeddingProvider
from .providers.eval import EvalProvider
from .providers.llm import GenerationConfig, LLMConfig, LLMProvider
from .providers.logging import LoggingDatabaseConnection, log_execution_to_db
from .providers.prompt import DefaultPromptProvider, PromptProvider
from .providers.vector_db import (
    VectorDBProvider,
    VectorEntry,
    VectorSearchResult,
)

__all__ = [
    "BasicDocument",
    "DefaultPromptProvider",
    "RAGPipelineOutput",
    "EmbeddingPipeline",
    "EvalPipeline",
    "IngestionPipeline",
    "RAGPipeline",
    "LoggingDatabaseConnection",
    "log_execution_to_db",
    "PromptProvider",
    "EvalProvider",
    "EmbeddingProvider",
    "GenerationConfig",
    "LLMConfig",
    "LLMProvider",
    "VectorSearchResult",
    "VectorEntry",
    "VectorDBProvider",
]



from typing import Any, Optional

from r2r.core import (
    EvalPipeline,
    LoggingDatabaseConnection,
    log_execution_to_db,
)


class BasicEvalPipeline(EvalPipeline):
    eval_providers = ["deepeval", "parea", "none"]

    def __init__(
        self,
        eval_config: dict,
        logging_connection: Optional[LoggingDatabaseConnection] = None,
        *args,
        **kwargs,
    ):
        frequency = eval_config["frequency"]
        super().__init__(frequency, logging_connection, *args, **kwargs)
        provider = eval_config["provider"]
        if provider not in self.eval_providers:
            raise ValueError(
                f"EvalProvider {provider} not supported in `BasicEvalPipeline`."
            )

        if provider == "deepeval":
            try:
                from r2r.eval import DeepEvalProvider
            except ImportError:
                raise ImportError(
                    "DeepEval is not installed. Please install it using `pip install deepeval`."
                )
            self.eval_provider = DeepEvalProvider(
                provider,
                eval_config.get("sampling_fraction", 1.0),
            )

        elif provider == "parea":
            try:
                from r2r.eval import PareaEvalProvider
            except ImportError:
                raise ImportError(
                    "Parea is not installed. Please install it using `pip install parea-ai`."
                )
            self.eval_provider = PareaEvalProvider(
                provider,
                eval_config.get("sampling_fraction", 1.0),
            )
        else:
            self.eval_provider = None

    @log_execution_to_db
    def evaluate(self, query: str, context: str, completion: str) -> Any:
        return self.eval_provider.evaluate(query, context, completion)



import asyncio
import json
import os
import subprocess
import tempfile
import time

import pytest

from r2r.client import R2RClient
from r2r.core.utils import generate_id_from_label


@pytest.fixture(scope="session", autouse=True)
def r2r_server():
    # Start the R2R server using poetry run uvicorn
    server_process = subprocess.Popen(
        [
            "poetry",
            "run",
            "uvicorn",
            "r2r.examples.basic.app:app",
            "--port=8010",
            "--workers=1",
        ]
    )

    # Wait for the server to start up (adjust the delay if needed)
    time.sleep(5)

    yield

    # Stop the R2R server after the tests complete
    server_process.terminate()
    server_process.wait()

    # Delete the local.sqlite file
    sqlite_file = "local.sqlite"
    if os.path.exists(sqlite_file):
        os.remove(sqlite_file)
    else:
        raise FileNotFoundError(f"Expected {sqlite_file} not found")


@pytest.fixture(scope="module")
def client():
    base_url = "http://localhost:8010"
    return R2RClient(base_url)


def test_upload_and_process_json(client):
    # Create a temporary JSON file
    json_data = {"key": "value"}
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".json", delete=False
    ) as temp_json:
        json.dump(json_data, temp_json)
        temp_json_path = temp_json.name

    try:
        # Upload and process the JSON file
        metadata = {"tags": ["json", "test"]}
        document_id = generate_id_from_label("test_json")
        upload_json_response = client.upload_and_process_file(
            document_id, temp_json_path, metadata, None
        )

        assert "processed and saved" in upload_json_response["message"]

        # Perform a search on the uploaded JSON file
        search_response = client.search(
            "value",
            5,
            filters={"document_id": document_id},
        )

        assert len(search_response) > 0

    finally:
        # Delete the temporary JSON file
        os.unlink(temp_json_path)

        # Delete the uploaded document
        client.filtered_deletion("document_id", document_id)


def test_upload_and_process_html(client):
    # Create a temporary HTML file
    html_content = (
        "<html><body><h1>Test HTML</h1><p>This is a test.</p></body></html>"
    )
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".html", delete=False
    ) as temp_html:
        temp_html.write(html_content)
        temp_html_path = temp_html.name

    try:
        # Upload and process the HTML file
        metadata = {"tags": ["html", "test"]}
        document_id = generate_id_from_label("test_html")
        upload_html_response = client.upload_and_process_file(
            document_id, temp_html_path, metadata, None
        )

        assert "processed and saved" in upload_html_response["message"]

        # Perform a search on the uploaded HTML file
        search_response = client.search(
            "test",
            5,
            filters={"document_id": document_id},
        )

        assert len(search_response) > 0

    finally:
        # Delete the temporary HTML file
        os.unlink(temp_html_path)

        # Delete the uploaded document
        client.filtered_deletion("document_id", document_id)


# TODO - Modify these so that they work deterministically
# def test_search(client):
#     search_response = client.search("test", 5)
#     assert len(search_response) > 0

# def test_filtered_search(client):
#     filtered_search_response = client.search(
#         "test", 5, filters={"tags": "bulk"}
#     )
#     assert len(filtered_search_response) > 0
#     for result in filtered_search_response:
#         assert "bulk" in result["metadata"]["tags"]

# def test_filtered_deletion(client):
#     document_id = generate_id_from_label("doc 2")
#     response = client.filtered_deletion("document_id", document_id)
#     assert response["deleted"] == 1

#     post_deletion_filtered_search_response = client.search(
#         "test", 5, filters={"tags": "bulk"}
#     )
#     assert len(post_deletion_filtered_search_response) == 0


def test_upload_and_process_pdf(client):
    current_file_directory = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(current_file_directory, "test.pdf")

    metadata = {"tags": ["example", "test"]}
    document_id = generate_id_from_label("pdf 1")
    upload_pdf_response = client.upload_and_process_file(
        document_id, file_path, metadata, None
    )
    assert "processed and saved" in upload_pdf_response["message"]

    pdf_filtered_search_response = client.search(
        "what is a cool physics equation?",
        5,
        filters={"document_id": document_id},
    )
    assert len(pdf_filtered_search_response) > 0


def test_rag_completion(client):
    document_id = generate_id_from_label("pdf 1")
    rag_response = client.rag_completion(
        "Are there any test documents?",
        5,
        filters={"document_id": document_id},
    )
    assert rag_response  # Add your own assertions here based on the expected response


@pytest.mark.skip(
    reason="RAG streaming functionality is commented out in the original file"
)
def test_stream_rag_completion(client):
    document_id = generate_id_from_label("pdf 1")

    async def stream_rag_completion():
        async for chunk in client.stream_rag_completion(
            "Are there any test documents?",
            5,
            filters={"document_id": document_id},
            generation_config={"stream": True},
        ):
            assert chunk  # Add your own assertions here based on the expected chunks

    asyncio.run(stream_rag_completion())


def test_get_logs(client):
    logs_response = client.get_logs()
    assert logs_response  # Add your own assertions here based on the expected logs


def test_get_logs_summary(client):
    logs_summary_response = client.get_logs_summary()
    assert logs_summary_response  # Add your own assertions here based on the expected logs summary



import asyncio
import glob
import os

import fire

from r2r.client import R2RClient
from r2r.core.utils import generate_id_from_label


class PDFChat:
    def __init__(self, base_url="http://localhost:8000", user_id=None):
        self.client = R2RClient(base_url)
        if not user_id:
            self.user_id = generate_id_from_label("user_id")
        self.titles = {
            "meditations.pdf": "Title: Meditations - Marcus Aurelius",
            # uncomment the following line to add more documents
            # "the_republic.pdf": "Title: The Republic - Plato",
        }

    def ingest(self):
        current_file_directory = os.path.dirname(os.path.abspath(__file__))
        data_path = os.path.join(current_file_directory, "..", "data")
        for file_path in glob.glob(os.path.join(data_path, "*.pdf")):
            file_name = file_path.split(os.path.sep)[-1]
            if file_name in self.titles:
                document_id = generate_id_from_label(file_path)
                metadata = {
                    "user_id": self.user_id,
                    "chunk_prefix": self.titles[file_name],
                }
                settings = {}
                upload_response = self.client.upload_and_process_file(
                    document_id, file_path, metadata, settings
                )
                print("Upload response = ", upload_response)

    def search(self, query):
        search_response = self.client.search(
            query,
            5,
            filters={"user_id": self.user_id},
        )
        for i, response in enumerate(search_response):
            text = response["metadata"]["text"]
            title, body = text.split("\n", 1)
            print(f"Result {i + 1}: {title}")
            print(body[:500])
            print("\n")

    def rag_completion(self, query, model="gpt-4-turbo-preview"):
        rag_response = self.client.rag_completion(
            query,
            5,
            filters={"user_id": self.user_id},
            generation_config={"model": model},
        )
        print("rag_response = ", rag_response)

    def rag_completion_streaming(self, query, model="gpt-4-turbo-preview"):
        async def stream_rag_completion():
            async for chunk in self.client.stream_rag_completion(
                query,
                5,
                filters={"user_id": self.user_id},
                generation_config={"stream": True, "model": model},
            ):
                print(chunk, end="", flush=True)

        asyncio.run(stream_rag_completion())

    def delete_document(self, document_path: str):
        document_id = generate_id_from_label(document_path)
        response = self.client.filtered_deletion("document_id", document_id)
        print("Deletion response = ", response)

    def get_logs(self):
        print("Fetching logs after all steps...")
        logs_response = self.client.get_logs()
        print(f"Logs response:\n{logs_response}\n")

    def get_logs_summary(self):
        print("Fetching logs summary after all steps...")
        logs_summary_response = self.client.get_logs_summary()
        print(f"Logs summary response:\n{logs_summary_response}\n")


if __name__ == "__main__":
    fire.Fire(PDFChat)
